#ifndef DUMPABLE_H_INCLUDED_C1D2A5B5
#define DUMPABLE_H_INCLUDED_C1D2A5B5

#include"RACE_LogfileOffset.h"
#include"RACE_Message_Block.h"

namespace RACE
{
  //##Documentation
  //## All Visitors/Semantics inherit from this interface
  class DumpableVisitor
  {
  public:
    //##Documentation
    //## make artificialy an interface
    virtual const char *name()=0;
  };
}

//##ModelId=3F72A5C5017C
//##Documentation
//## This is the Interface for classes dumped into a Logfile
class RACE_Dumpable: public RACE_Message_Block
{
  public:
    //##ModelId=3F72A5C6014B
    RACE_Dumpable(size_t size):RACE_Message_Block(size) {}
    //##Documentation
    //## for read side 
    RACE_Dumpable(RACE::DataBlock *db,bool frag,RACE_LogfileOffset lo):RACE_Message_Block(db,frag),offset_(lo) {}
    //##ModelId=3F72A5C602E0
    virtual ~RACE_Dumpable();
    //##ModelId=3F72A5C60351
    //##Documentation
    //## specifies if the Data is not so important,
    //## thus it can be discarded if too busy
    virtual bool IsDiscardable();

    //##ModelId=3F72A5C603CB
    //##Documentation
    //## When the offset in the Logfile is known,
    //## the Caller must be informed with the offset.
    virtual void InformOffset(RACE_LogfileOffset offset);
    //Documentation
    //## Set the Offset - mainly intended for the read side
    //void SetOffset(const RACE_LogfileOffset& offset) { offset_ = offset; }
    //##Documentation
    //## get the Offset - mainly intended for the read side
    RACE_LogfileOffset GetOffset() { return offset_; }
    //##ModelId=3F72A5C70173
    //##Documentation
    //## If a Dumpable is discarded, potentaly some work has to be done,
    //## for instance
    //##  * a snapshot waits for it,
    //##  * a statistic, on which a strategy can calculate
    //## if the next Dumpable should be despensable
    virtual void Discard();

    //##Documentation
    //## Hook for the Reading Semantik
    virtual void CallVisitor(RACE::DumpableVisitor *visitor);
 protected:
    //##ModelId=3F72A5C701EB
    RACE_Message_Block *CloneManagementData() const;
    //##ModelId=3F72A5C70296
    RACE_Dumpable():RACE_Message_Block() {}
    RACE_LogfileOffset offset_;
};



#endif /* DUMPABLE_H_INCLUDED_C1D2A5B5 */

