#ifndef RACELOGINDEXPOSITIONER_H
#define RACELOGINDEXPOSITIONER_H
#include"RACE_LogIndexLayout.h"
#include<stdio.h>

//##ModelId=3FB37BB50355
//##Documentation
//##  Reads the indexfile and maps the starting time to offset of the nearest
//##  older Snapshot
class RACE_LogIndexPositioner
{
public:
    //##ModelId=3FB37BB50356
  class RangeError
  {
  public:
        //##ModelId=3FB37BB60329
    RangeError(unsigned long index,unsigned long max)
      :m_index(index),m_max(max) {}
        //##ModelId=3FB37BB60335
    unsigned long GetIndex() { return m_index; }
        //##ModelId=3FB37BB60336
    unsigned long GetMax()   { return m_max; }
        //##ModelId=3FB37BB60337
    const char *What()
    {
      snprintf(m_errstr,80,"RACE_LogIndexPositioner: RangeError %lu (of %lu)",m_index,m_max);
      return m_errstr;
    }
  private:
        //##ModelId=3FB37BB60326
    unsigned long m_index;
        //##ModelId=3FB37BB60327
    unsigned long m_max;
        //##ModelId=3FB37BB60328
    char m_errstr[80];
  };
    //##ModelId=3FB37BB60345
  RACE_LogIndexPositioner():mp_entries(0),m_count(0) {}
    //##ModelId=3FB37BB60346
  RACE_LogIndexPositioner(const char *filename);

  //##ModelId=3FB37BB60348
  //##Documentation
  //##  Opens the Indexfile
  void Open(const char *filename);
    //##ModelId=3FB37BB6034A
  ~RACE_LogIndexPositioner();

  //##ModelId=3FB37BB6034B
  //##Documentation
  //##  Prints the Index on stdout
  void Print();

  //##ModelId=3FB37BB6034C
  //##Documentation
  //##  gets the number of Entries in the Indexfile
  unsigned long GetCount() { return m_count; }

  //##ModelId=3FB37BB6034D
  //##Documentation
  //##  gets the i-th entry
  RACE_LogIndexLayout& operator[](unsigned long i) throw (RACE_LogIndexPositioner::RangeError);

  //##ModelId=3FB37BB60355
  //##Documentation
  //## Gets the last Index before time wst
  unsigned long GetLastIndex(time_t wst);

  //##ModelId=3FB37BB60357
  //##Documentation
  //##  Converts a string of the Form hh:mm:ss into a Timestamp
  //##  The Date is fetched from the first entry and the time is overwritten
  //##  with the parsed time from the string
  time_t StrToTime(const char *hhmmss);

  //##ModelId=3FB37BB60359
  //##Documentation
  //##  gets offset from corresponding Timestamp
  RACE_LogfileOffset Get(time_t wst) { return (*this)[GetLastIndex(wst)].offset; }

    //##ModelId=3FB37BB6035B
  //##Documentation
  //## Get the last offset in the Index
  RACE_LogfileOffset GetLastOffset() { return (*this)[m_count-1].offset; }

    //##ModelId=3FB37BB6035C
  //##Documentation
  //## checks if the index is opened and Entries in the Index
  bool IsEmpty() { return m_fd<0 || m_count<=0; }
protected:
    //##ModelId=3FB37BB60338
  int m_fd;
    //##ModelId=3EBA5D420029
  RACE_LogIndexLayout *mp_entries;
    //##ModelId=3FB37BB6033E
  unsigned long m_count;
};
#endif

