#ifndef RACE_MESSAGE_BLOCK
#define RACE_MESSAGE_BLOCK
#include<sys/types.h>
#include<ace/Synch_T.h>
#include<utility/errorhandling.h>
#include"RACE_DataBlock.h"


//##ModelId=3F72A5C501D0
//##Documentation
//##  RACE_Message_Block
class RACE_Message_Block
{

public:
    //##ModelId=3F72A5C902FE
  //##Documentation
  //## generates a empty Message_Block with size len
  RACE_Message_Block(size_t len);

  //##Documentation
  //## Generates a Messageblock with a existent Data_Block
  //## Important for the read side
  RACE_Message_Block(RACE::DataBlock *db,bool frag);

    //##ModelId=3F72A5C903DE
  virtual ~RACE_Message_Block();
  // = Operations on Message Data
//##ModelId=3F72A5CA002F
//##Documentation
//## Resets the Message Block data to contain nothing, i.e., sets the
//## read and write pointers to align with the base.
  void reset (void) {  rd_ptr_=0;  wr_ptr_=0; }

//##ModelId=3F72A5CA00E5
//##Documentation
//## Get message data.
  char *base (void) const { return (char *)(db->begin()); }
//##ModelId=3F72A5CA019D
//##Documentation
//## Return a pointer to 1 past the end of the allocated data in a message.
  char *end (void) const  { return (char *)(db->begin())+db->length(); }

    //##ModelId=3F72A5CA0255
  char *first_base(void) { return (char *)(first_->begin()); }
//##ModelId=3F72A5CA030C
//##Documentation
//##  Copies <n> bytes from <buf> into the Message_Block starting at
//##  the <wr_ptr> offset.  Return 0 and increment <wr_ptr> by <n> if
//##  the method succeeds.  Returns -1 if the size of the message is
//##  too small, i.e., for this to work correct, <end> must be >=
//##  <wr_ptr>.

  int copy (const char *buf, size_t n) throw (BufferOverflow);

//##ModelId=3F72A5CB0093
//##Documentation
//##  Get the read pointer.
  char *rd_ptr (void) const { return base()+rd_ptr_; }
//##ModelId=3F72A5CB0149
//##Documentation
//##  Set the read pointer to <ptr>.
  void rd_ptr (char *ptr) { rd_ptr_=ptr-base(); }
//##ModelId=3F72A5CB0220
//##Documentation
//##  Set the read pointer ahead <n> bytes.
  void rd_ptr (size_t n)  { rd_ptr_ +=n; }

//##ModelId=3F72A5CB02F8
//##Documentation
//##  Get the write pointer.
  char *wr_ptr (void) const { return base()+wr_ptr_; }
//##ModelId=3F72A5CB03B3
//##Documentation
//##  Set the write pointer to <ptr>.
  void wr_ptr (char *ptr) throw (BufferOverflow)
    {
      size_t neu = ptr-base();
      if(neu>db->length()) db->grow(neu - db->length());
      wr_ptr_=neu;
    }
//##ModelId=3F72A5CC00A1
//##Documentation
//##  Set the write pointer ahead <n> bytes.
//##  Returns the OLD ptr
  char *wr_ptr (size_t n) throw (BufferOverflow)
  {
    char *ret=wr_ptr();
    if(wr_ptr_+n>db->length()) db->grow(wr_ptr_+n - db->length());
    wr_ptr_+=n;
    return ret;
  }

  //##Documentation
  //## aligns the wrptr to a boundary
  //## returns the new ptr
  char *wrptr_align(short potenz2);

//##ModelId=3F72A5CC0174
//##Documentation
//## Get the number of bytes in the Message Block
  size_t size(void) const { return db->size(); }
//##ModelId=3F72A5CC022E
//##Documentation
//## Get the length of the message
//## CAUTION: Different from ACE: its the real length,
//##          independed from the rd_ptr
  size_t length(void) const { return db->length(); }
//##ModelId=3F72A5CC02EB
//##Documentation
//## remaining : the count of unconsumed bytes dependend from the rd_ptr
  size_t remaining(void) const { return wr_ptr_-rd_ptr_; }

  /// Set the length of the message
  //void length (size_t n);

  //##Documentation
  //## returns the length of free memory in the current Datablock
  size_t freespace() { return db->freespace(); }
//##ModelId=3F72A5CC03B3
//##Documentation
//## Get the length of all DataBlocks in the <Message_Block>
  size_t total_length(void) const;

//##ModelId=3F72A5CD0083
//##Documentation
//## Get the total number of bytes in all DataBlocks in the  <Message_Block>
  size_t total_size(void) const;

//##ModelId=3F72A5CD013D
//##Documentation
//## allocates a new block for growing messageblocks
  void expand();
  //##Documentation
  //## adds a new datablock
  void addblock(RACE::DataBlock *block);

    //##ModelId=3F72A5CD0178
  bool IsFragmented() { return (first_!=NULL && first_->next !=NULL); }
//##ModelId=3F72A5CD01B9
//##Documentation
//##iteration
  void begin()   { db=first_; if(db) {reset_current();} }
    //##ModelId=3F72A5CD01F3
  bool is_done() { return db==NULL; }
    //##ModelId=3F72A5CD022E
  void advance() { db=db->next; if(db) {reset_current();} }

  // = Operations on the whole Message

  /// Return an exact "deep copy" of the message, i.e., create fresh
  /// new copies of all the Data_Blocks and continuations.
  // virtual RACE_Message_Block *clone (Message_Flags mask = 0) const;

//##ModelId=3F72A5CD0269
//##Documentation
//## Return a "shallow" copy that increments our reference count by 1.
  RACE_Message_Block *duplicate(void) const;

//##ModelId=3F72A5CD0326
//##Documentation
//## Return a deep copy
  RACE_Message_Block *clone(void) const;

    //##ModelId=3F72A5CD03E3
  void release() { delete this; }
//##ModelId=3F72A5CE0035
//##Documentation
//## Get link to next message.
  RACE_Message_Block *next (void) const { return next_; }
//##ModelId=3F72A5CE00F3
//##Documentation
//## Set link to next message.
  void next(RACE_Message_Block *n)      { next_ = n; }

    // = Pointer to the <Message_Block> directly behind in the <ACE_Message_Queue>.
//##ModelId=3F72A5CE01CA
//##Documentation
//## Get link to prev message.
  RACE_Message_Block *prev(void) const  { return prev_; }
//##ModelId=3F72A5CE0287
//##Documentation
//## Set link to prev message.
  void prev(RACE_Message_Block *p)      { prev_=p; }
private:
    //##ModelId=3F72A5C900E9
  RACE::DataBlock *db;  //the current Messageblock
    //##ModelId=3F72A5C900F6
  RACE::DataBlock *first_;
    //##ModelId=3F72A5C90102
  size_t block_size_;
    //##ModelId=3F72A5C901A2
  size_t rd_ptr_; // Relative
    //##ModelId=3F72A5C9023D
  size_t wr_ptr_; // Relative

protected:
//##ModelId=3F72A5CE035D
//##Documentation
//##  The Cloning of derivations is only posible with a virtual clone
//##  operation. It is called from duplicate()

  virtual RACE_Message_Block *CloneManagementData() const;

//##ModelId=3F72A5CE03B8
//##Documentation
//##  The default constuctur is only for cloning the management data

  RACE_Message_Block()
  {
    first_ = db = (RACE::DataBlock *)0;
    rd_ptr_=0;
    wr_ptr_=0;
    next_=0;
    prev_=0;
  }
//##ModelId=3F72A5CF002A
//##Documentation
//## set current data
  void reset_current()
  {
    rd_ptr_=0;
    wr_ptr_=db->length();
  }


    //##ModelId=3F72A5C902DC
  RACE_Message_Block *next_;
    //##ModelId=3F72A5C902EE
  RACE_Message_Block *prev_;
};


#endif

