#ifndef RACE_MESSAGE_QUEUE_T_C
#define RACE_MESSAGE_QUEUE_T_C

#include "RACE_Message_Queue_T.h"
#include <utility/envdebug.h>
#include <utility/errorhandling.h>

#if !defined (ACE_LACKS_PRAGMA_ONCE)
# pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

static envdebug *race_mqt_debug = new envdebug("RACE_Message_Queue");

//##ModelId=3F72A5CF00CA
template <ACE_SYNCH_DECL>
RACE_Message_Queue<ACE_SYNCH_USE>::RACE_Message_Queue(RACE_QueueLevelStrategy *qls)
 : not_empty_cond_ (this->lock_),
   not_full_cond_ (this->lock_),
   qls_(qls)
{
  race_mqt_debug->print("RACE_Message_Queue<ACE_SYNCH_USE>::RACE_Message_Queue");
  RACE_COND_FAIL(qls_==0,"RACE_Message_Queue<ACE_SYNCH_USE>::RACE_Message_Queue: no QueueLevelStrategy");
  int r=this->open();
  RACE_COND_FAIL(r==-1,"RACE_Message_Queue<ACE_SYNCH_USE>::open failed");
}

//##ModelId=3F72A5CF00DC
template <ACE_SYNCH_DECL>
RACE_Message_Queue<ACE_SYNCH_USE>::~RACE_Message_Queue(void)
{
  race_mqt_debug->print("RACE_Message_Queue<ACE_SYNCH_USE>::~RACE_Message_Queue");
  if (this->head_ != 0 && this->close() == -1)
  {
    output::fail("RACE_Message_Queue<ACE_SYNCH_USE>::close failed");
  }
}

//##ModelId=3F72A5CF00D1
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::open()
{
  race_mqt_debug->print("RACE_Message_Queue<ACE_SYNCH_USE>::open");
  this->tail_ = 0;
  this->head_ = 0;
  return 0;
}

//##ModelId=3F72A5CF00D4
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::close(void)
{
  race_mqt_debug->print("RACE_Message_Queue<ACE_SYNCH_USE>::close");
  ACE_GUARD_RETURN (ACE_SYNCH_MUTEX_T, ace_mon, this->lock_, -1);
  int res=0;
  // Free up the remaining messages on the queue.
  for (this->tail_ = 0; this->head_ != 0; )
    {
      this->qls_->remove(this->head_->total_size(),
                         this->head_->total_length());

      RACE_Message_Block *temp = this->head_;
      this->head_ = this->head_->next ();

      // Make sure to use <release> rather than <delete> since this is
      // reference counted.
      temp->release();
    }

  return res;
}

//##ModelId=3F72A5CF0139
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::wait_not_full_cond(ACE_Time_Value *timeout)
{
  int result = 0;

  // Wait while the queue is full.
  while (this->qls_->is_full())
  {
    if (this->not_full_cond_.wait (timeout) == -1)
    {
      if (errno == ETIME)
        errno = EWOULDBLOCK;
      result = -1;
      break;
    }
  }
  return result;
}

//##ModelId=3F72A5CF0141
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::wait_not_empty_cond(ACE_Time_Value *timeout)
{
  int result = 0;
  // Wait while the queue is empty

  while (this->qls_->is_empty())
  {
    if (this->not_empty_cond_.wait (timeout) == -1)
    {
      if (errno == ETIME)
        errno = EWOULDBLOCK;
      result = -1;
      break;
    }
  }
  return result;
}

//##ModelId=3F72A5CF014B
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::signal_enqueue_waiters (void)
{
  if (this->not_full_cond_.signal () != 0) { return -1; }
  return 0;
}

//##ModelId=3F72A5CF0153
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::signal_dequeue_waiters (void)
{
  // Tell any blocked threads that the queue has a new item!
  if (this->not_empty_cond_.signal () != 0) { return -1; }
  return 0;
}

// Actually put the node at the end (no locking so must be called with
// locks held).

//##ModelId=3F72A5CF0127
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::enqueue_tail_i(RACE_Message_Block *new_item)
{
  race_mqt_debug->print("ACE_Message_Queue<ACE_SYNCH_USE>::enqueue_tail_i");

  if (new_item == 0)
    return -1;

  // List was empty, so build a new one.
  if (this->tail_ == 0)
    {
      this->head_ = new_item;
      this->tail_ = new_item;
      new_item->next(0);
      new_item->prev(0);
    }
  // Link at the end.
  else
    {
      new_item->next(0);
      this->tail_->next(new_item);
      new_item->prev(this->tail_);
      this->tail_ = new_item;
    }

  // Make sure to count all the bytes in a composite message!!!
  this->qls_->add(new_item->total_size(),new_item->total_length());

  if (this->signal_dequeue_waiters() == -1)
    return -1;
  else
    return this->qls_->count();
}

// Actually get the first ACE_Message_Block (no locking, so must be
// called with locks held).  This method assumes that the queue has at
// least one item in it when it is called.

//##ModelId=3F72A5CF0130
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::dequeue_head_i(RACE_Message_Block *&first_item)
{
  if(this->head_ ==0)
  {
    RACE_ERROR("ACE_Message_Queue<ACE_SYNCH_USE>::dequeue_head_i: Attempting to dequeue from empty queue");
    return -1;
  }
  race_mqt_debug->print("ACE_Message_Queue<ACE_SYNCH_USE>::dequeue_head_i");
  first_item = this->head_;
  this->head_ = this->head_->next ();

  if (this->head_ == 0)
    this->tail_ = 0;
  else
    // The prev pointer of the first message block has to point to
    // NULL...
    this->head_->prev (0);

  // Subtract off all of the bytes associated with this message.
  this->qls_->remove(first_item->total_size(),first_item->total_length());

  if (this->qls_->count() == 0 && this->head_ == this->tail_)
  {
    this->head_ = this->tail_ = 0;
  }

  // Only signal enqueueing threads if we've fallen below the low
  // water mark.
  if (this->qls_->can_enqueue()
      && this->signal_enqueue_waiters() == -1)
    return -1;
  else
    return this->qls_->count();
}

// Check if queue is empty (holds locks).

//##ModelId=3F72A5CF0102
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::is_empty(void)
{
  race_mqt_debug->print("RACE_Message_Queue<ACE_SYNCH_USE>::is_empty");
  ACE_GUARD_RETURN (ACE_SYNCH_MUTEX_T, ace_mon, this->lock_, -1);

  return this->qls_->is_empty();
}

// Check if queue is full (holds locks).


//##ModelId=3E79BAF5025E
template <ACE_SYNCH_DECL> ACE_INLINE
//##ModelId=3F72A5CF0109
 int
RACE_Message_Queue<ACE_SYNCH_USE>::is_full(void)
{
  race_mqt_debug->print("RACE_Message_Queue<ACE_SYNCH_USE>::is_full");
  ACE_GUARD_RETURN (ACE_SYNCH_MUTEX_T, ace_mon, this->lock_, -1);

  return this->qls_->is_full();
}

//##ModelId=3F72A5CF00E3
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::enqueue(RACE_Message_Block *new_item)
{
  race_mqt_debug->print("ACE_Message_Queue<ACE_SYNCH_USE>::enqueue");
  ACE_GUARD_RETURN (ACE_SYNCH_MUTEX_T, ace_mon, this->lock_, -1);

  if (this->wait_not_full_cond(this->qls_->GetCurrentEnqueueTimeout()) == -1)
    return -1;

  int queue_count = this->enqueue_tail_i(new_item);

  if (queue_count == -1)
    return -1;

  return queue_count;
}

//##ModelId=3F72A5CF00EB
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::enqueue_force(RACE_Message_Block *new_item)
{
  race_mqt_debug->print("ACE_Message_Queue<ACE_SYNCH_USE>::enqueue");
  ACE_GUARD_RETURN (ACE_SYNCH_MUTEX_T, ace_mon, this->lock_, -1);

  int queue_count = this->enqueue_tail_i(new_item);

  if (queue_count == -1)
    return -1;

  return queue_count;
}

//##ModelId=3F72A5CF00F3
template <ACE_SYNCH_DECL> int
RACE_Message_Queue<ACE_SYNCH_USE>::dequeue(RACE_Message_Block *&first_item)
{
  race_mqt_debug->print("ACE_Message_Queue<ACE_SYNCH_USE>::dequeue");
  ACE_GUARD_RETURN (ACE_SYNCH_MUTEX_T, ace_mon, this->lock_, -1);

  if(this->wait_not_empty_cond(this->qls_->GetCurrentDequeueTimeout()) == -1) return -1;
  return dequeue_head_i(first_item);
}

//##ModelId=3F72A5CF0110
template <ACE_SYNCH_DECL> size_t
RACE_Message_Queue<ACE_SYNCH_USE>::message_bytes(void)
{
  return qls_->size_sum();
}

//##ModelId=3F72A5CF0119
template <ACE_SYNCH_DECL>
size_t RACE_Message_Queue<ACE_SYNCH_USE>::message_length(void)
{
  return qls_->length_sum();
}

//##ModelId=3F72A5CF0120
template <ACE_SYNCH_DECL>
size_t RACE_Message_Queue<ACE_SYNCH_USE>::message_count(void)
{
  return qls_->count();
}
#endif /* RACE_MESSAGE_QUEUE_T_C */

