#include "RACE_QueueLevelStrat.h"
#include <utility/envdebug.h>
#include <utility/errorhandling.h>

//##ModelId=3F72A5D1013F
static envdebug *debug = new envdebug("RACE_QueueLevelStrategy");

//##ModelId=3F72A5D203D4
void RACE_QueueLevelStrategy::add(size_t cur_size,size_t cur_length)
{
  cur_count_++;
  cur_lengthsum_ += cur_length;
  cur_sizesum_   += cur_size;
  debug->print("RACE_QueueLevelStrategy::add after count=%d length=%d bytes size=%d bytes",cur_count_,cur_lengthsum_,cur_sizesum_);
  if(is_critical_watermark_reached())
  {
    if(!critical_watermark_triggered)
    {
      critical_watermark_alarm();
      critical_watermark_triggered=true;
    }
  }
  else
  {
    // Alarm is over
    critical_watermark_triggered=false;
  }
}

//##ModelId=3F72A5D3032C
void RACE_QueueLevelStrategy::remove(size_t cur_size,size_t cur_length)
{
  RACE_COND_FAIL(cur_count_<1,
                    "RACE_QueueLevelStrategy::remove: Queuesize<0");
  cur_count_--;  // assert cur_count_ >=0
  cur_lengthsum_ -= cur_length;
  cur_sizesum_   -= cur_size;
}

//##ModelId=3F72A5D500F3
bool RACE_QueueLevelStrategy::is_critical_watermark_reached()
{
  return false;
}

//##ModelId=3F72A5D50166
const char *RACE_QueueLevelStrategy::GetQueueName()
{
  return "";
}

//##ModelId=3F72A5D501D8
void RACE_QueueLevelStrategy::critical_watermark_alarm()
{
  RACE_WARN("Critical Watermark in Queue %s is reached",GetQueueName());
}

//##ModelId=3F72A5D50252
ACE_Time_Value *RACE_QueueLevelStrategy::GetCurrentEnqueueTimeout()
{
  return 0;
}

//##ModelId=3F72A5D502C6
ACE_Time_Value *RACE_QueueLevelStrategy::GetCurrentDequeueTimeout()
{
  return 0;
}

//##ModelId=3F72A5D50339
RACE_QueueLevelStrategy::~RACE_QueueLevelStrategy() {}

//##ModelId=3F72A5D1010F
bool RACE_QLS_Count::can_enqueue()
{
  return cur_count_ < max_count_;
}

//##ModelId=3F72A5D10112
bool RACE_QLS_Count::is_full()
{
  return cur_count_ >= max_count_;
  // >: defensive programming
}

//##ModelId=3F72A5D1012C
bool RACE_QLS_MessageSizeSum::can_enqueue()
{
  return cur_sizesum_ <= this->low_watermark_;
}

//##ModelId=3F72A5D1012E
bool RACE_QLS_MessageSizeSum::is_full()
{
  return cur_sizesum_ > high_watermark_;
}


//##ModelId=3F72A5D10114
RACE_QLS_Count::~RACE_QLS_Count()
{
}

//##ModelId=3F72A5D1012A
RACE_QLS_MessageSizeSum::~RACE_QLS_MessageSizeSum()
{
}

