#ifndef RACE_QUEUELEVELSTRAT_H
#define RACE_QUEUELEVELSTRAT_H
#include<sys/types.h>
#include<ace/Time_Value.h>

//##ModelId=3F72A5C50218
//##Documentation
//## This is the strategy class which controls the RACE_Message_Queue.
class RACE_QueueLevelStrategy
{
public:
    //##ModelId=3F72A5D202F0
  RACE_QueueLevelStrategy():cur_sizesum_(0),cur_lengthsum_(0), cur_count_(0),
    critical_watermark_triggered(false) {}
    //##ModelId=3F72A5D20361
  void reset() { cur_sizesum_=0;cur_lengthsum_=0;cur_count_=0; }
    //##ModelId=3F72A5D203D4
  void add(size_t cur_size,size_t cur_length);
    //##ModelId=3F72A5D3032C
  void remove(size_t cur_size,size_t cur_length);
    //##ModelId=3F72A5D4021D
  size_t count()     { return cur_count_; }
    //##ModelId=3F72A5D40291
  size_t size_sum()  { return cur_sizesum_; }
    //##ModelId=3F72A5D40303
  size_t length_sum(){ return cur_lengthsum_; }

    //##ModelId=3F72A5D4037B
  bool is_empty() { return cur_count_ == 0; }
    //##ModelId=3F72A5D50007
  virtual bool can_enqueue()=0;
    //##ModelId=3F72A5D5007A
  virtual bool is_full() = 0;

    //##ModelId=3F72A5D500F3
  virtual bool is_critical_watermark_reached();

    //##ModelId=3F72A5D50166
  //##Documentation
  //## Get the QueueName for Debugging and Warning purpose
  virtual const char *GetQueueName();

    //##ModelId=3F72A5D501D8
  //##Documentation
  //## A callback function which is called in add()
  //## when is_critical_watermark_reached()
  //## Derived stategies can overwrite this
  virtual void critical_watermark_alarm();


    //##ModelId=3F72A5D50252
  //##Documentation
  //## Class ACE_Condition requires an absolute time.
  //## The current absolute time is calculated here.
  //## 0 for no timeout
  //## Default is 0
  virtual ACE_Time_Value *GetCurrentEnqueueTimeout();

    //##ModelId=3F72A5D502C6
  //##Documentation
  //## Class ACE_Condition requires an absolute time.
  //## The current absolute time is calculated here.
  //## 0 for no timeout
  //## Default is 0
  virtual ACE_Time_Value *GetCurrentDequeueTimeout();

    //##ModelId=3F72A5D50339
  virtual ~RACE_QueueLevelStrategy();
protected:
//##ModelId=3F72A5D10235
//##Documentation
//## Sum of Message sizes bytes in the queue.
  size_t cur_sizesum_;

//##ModelId=3F72A5D10361
//##Documentation
//## Current length of messages in the queue.
  size_t cur_lengthsum_;

//##ModelId=3F72A5D2009F
//##Documentation
//## Current number of messages in the queue.
  size_t cur_count_;


    //##ModelId=3F72A5D201C7
  bool critical_watermark_triggered;
};

//##ModelId=3F72A5C50200
class RACE_QLS_Count :public RACE_QueueLevelStrategy
{
public:
    //##ModelId=3F72A5D1010D
  RACE_QLS_Count(size_t max_count):max_count_(max_count) {}
    //##ModelId=3F72A5D1010F
  virtual bool can_enqueue();
    //##ModelId=3F72A5D10112
  virtual bool is_full();
    //##ModelId=3F72A5D10114
  virtual ~RACE_QLS_Count();
protected:
    //##ModelId=3F72A5D1010A
  size_t max_count_;
};

//##ModelId=3F72A5C5020A
//##Documentation
//##   The <high_water_mark>
//##  determines how many bytes can be stored in a queue before it's
//##  considered "full."  Supplier threads must block until the queue
//##  is no longer full.  The <low_water_mark> determines how many
//##  bytes must be in the queue before supplier threads are allowed to
//##  enqueue additional <RACE_Message_Block>s.  By default, the
//##  <high_water_mark> equals the <low_water_mark>, which means that
//##  suppliers will be able to enqueue new messages as soon as a
//##  consumer removes any message from the queue.  Making the
//##  <low_water_mark> smaller than the <high_water_mark> forces
//##  consumers to drain more messages from the queue before suppliers
//##  can enqueue new messages, which can minimize the "silly window
//##  syndrome."

class RACE_QLS_MessageSizeSum : public RACE_QueueLevelStrategy
{
 public:
    //##ModelId=3F72A5D10126
  RACE_QLS_MessageSizeSum(size_t low_watermark,size_t high_watermark)
    :low_watermark_(low_watermark),high_watermark_(high_watermark)
    {
    }
    //##ModelId=3F72A5D1012A
  virtual ~RACE_QLS_MessageSizeSum();
    //##ModelId=3F72A5D1012C
  virtual bool can_enqueue();
    //##ModelId=3F72A5D1012E
  virtual bool is_full();
 protected:
    //##ModelId=3F72A5D10121
  size_t low_watermark_;
    //##ModelId=3F72A5D10124
  size_t high_watermark_;
};
#endif

