#include "RACE_SnapshotTrigger.h"
#include <utility/envdebug.h>
#include <utility/errorhandling.h>

static envdebug *s_debug = new envdebug("RACE_SnapshotTrigger");

int RACE_SnapshotCreator::GetPendingTotal()
{
  return ST_UNKOWN;
}

//##ModelId=3F16A48D004E
int RACE_SnapshotCreator::GetPendingForYou()
{
  return ST_UNKOWN;
}

//##ModelId=3F16A48E03C6
RACE_SnapshotCreator::~RACE_SnapshotCreator()
{
}

//##ModelId=3F16A48F0104
RACE_SnapshotQueue::Entry::~Entry()
{
}

//##ModelId=3F16A48F012B
RACE_SnapshotQueue::~RACE_SnapshotQueue()
{
  s_debug->print("RACE_SnapshotQueue::~RACE_SnapshotQueue()");
  Entry* h;
  while(mp_first!=NULL)
  {
    h=mp_first;
    mp_first = mp_first->mp_next;
    delete h;
  }
}


//##ModelId=3F16A48F012E
void RACE_SnapshotQueue::GetCountAndSum(int& count,int& sumpending)
{
  count=0;
  sumpending=0;
  for(Entry *a=mp_first;a!=NULL;a=a->mp_next)
  {
    sumpending+=a->GetPending();
    count++;
  }
}

//##ModelId=3F16A48F0144
void RACE_SnapshotQueue::enqueue(RACE_SnapshotQueue::Entry *e)
{
  *mpp_last = e;
  mpp_last=&((*mpp_last)->mp_next);
}

//##ModelId=3F16A48F0158
RACE_SnapshotQueue::Entry *RACE_SnapshotQueue::dequeue()
{
  Entry *ret = mp_first;
  mp_first=mp_first->mp_next;
  if(mp_first==NULL)
  {
    mpp_last=&mp_first;
  }
  ret->mp_next=NULL;
  return ret;
}

//##ModelId=3F16A48F015D
RACE_SnapshotQueue::Entry *RACE_SnapshotQueue::GetEntry(time_t timestamp)
{
  Entry *cur=mp_first;
  while(cur!=NULL && cur->GetTimestamp() < timestamp)
  {
    cur=cur->mp_next;
  }
  if(cur!=NULL && cur->GetTimestamp() == timestamp)
  {
    return cur;
  }
  else
  {
    return NULL;
  }
}

/*---------------------- RACE_SnapshotTrigger --------------------*/
RACE_SnapshotTrigger::RACE_SnapshotTrigger(RACE_SnapshotCreator *sc,const char *description,RACE_SnapshotQueue *sq):mp_snapcreator(sc),m_description(description),
                    m_lastTimestamp(0),m_queue(sq)
{
  if(sq==NULL)
  {
    s_debug->print("RACE_SnapshotTrigger::RACE_SnapshotTrigger(%s): Using Default Queue",description);
    m_queue = new RACE_SnapshotQueue();
    RACE_COND_FAIL(m_queue==NULL,"%s:%d: No Memory ",__FILE__,__LINE__);
  }
}

//##ModelId=3F16A48F01CD
RACE_SnapshotTrigger::~RACE_SnapshotTrigger()
{
  s_debug->print("RACE_SnapshotTrigger::~RACE_SnapshotTrigger()");
  delete m_queue;
}



//##ModelId=3EFAEA1202AE
void RACE_SnapshotTrigger::PrepareSnapshot(time_t timestamp)
{
  int inothers=0;
  int count=0;
  int pending;
  int pendingforme;
  ACE_GUARD(ACE_Thread_Mutex, ace_mon, this->m_mutex);

  /*
   * The order of timestamps must be monoton
   * Multiple Perparations of the same snapshots occur in the
   * case of collecting WholeSnapshots
   */
  RACE_COND_FAIL(timestamp < m_lastTimestamp,
                   "RACE_SnapshotTrigger::PrepareSnapshot(%s,timestamp=%ld)<%ld",
                   m_description,timestamp,m_lastTimestamp);

  if(timestamp==m_lastTimestamp)
  {
    s_debug->print("RACE_SnapshotTrigger::PrepareSnapshot(%s,timestamp=%ld) already prepared",m_description,timestamp);
    return;
  }
  m_lastTimestamp = timestamp;

  m_queue->GetCountAndSum(count,inothers);

  if(mp_snapcreator->GetPendingForYou()==ST_UNKOWN)
  {
    /*
     * We know how many Dumpables pending total, but not for the current
     * Snapshot
     * It is in the case of OnEventSnapshots
     */
    pending      = mp_snapcreator->GetPendingTotal();
    pendingforme = pending - inothers;
    s_debug->print("RACE_SnapshotTrigger::PrepareSnapshot(%s %ld)-TN %d Dumpables for %d Snapshots pending, %d for current",
                   m_description,timestamp,pending,count,pendingforme);
  }
  else
  {
    /*
     * We know the number of pending Dumpables for the Current Snapshot
     */
    pendingforme = mp_snapcreator->GetPendingForYou();
    pending = pendingforme+inothers;
    s_debug->print("RACE_SnapshotTrigger::PrepareSnapshot(%s %ld)-CN %d Dumpables for %d Snapshots pending, %d for current",
   m_description,timestamp,pending,count,pendingforme);
  }


  m_queue->enqueue(mp_snapcreator->GenNewEntry(timestamp,pendingforme));

  if(pending<=0)  // < defensiv
  {
    s_debug->print("RACE_SnapshotTrigger::PrepareSnapshot(%s): no clones pending - logging",m_description);
    RACE_SnapshotQueue::Entry* tolog = m_queue->dequeue();
    tolog->Log();
    delete tolog;
  }
}

/*
 * Older Implementation
 * It is the same as InformInsert(GetTopTimestamp())
 *
 */
//##ModelId=3EFAEA1202B6
void RACE_SnapshotTrigger::InformInsert()
{
  int c;

  ACE_GUARD(ACE_Thread_Mutex, ace_mon, this->m_mutex);
  if(! m_queue->IsEmpty())
  {
    m_queue->DecrTopPending();
    s_debug->print("RACE_SnapshotTrigger::InformInsert(%s) %d pending",
                   m_description,m_queue->GetTopPending());
    /*
     * If the next pending snapshots don't wait for pending Dumpables
     * Log them too
     */
    c=1;
    while(!m_queue->IsEmpty() && m_queue->GetTopPending() == 0)
      {
      s_debug->print("RACE_SnapshotTrigger::InformInsert(%s): triggering %ld (%d in batch)",
                     m_description,m_queue->GetTopTimestamp(),c);
      RACE_SnapshotQueue::Entry* tolog = m_queue->dequeue();
      tolog->Log();
      delete tolog;
      c++;
      }
  }
}

//##ModelId=3F16A48F01D4
void RACE_SnapshotTrigger::InformInsert(time_t timestamp)
{
  int c;

  ACE_GUARD(ACE_Thread_Mutex, ace_mon, this->m_mutex);
  RACE_SnapshotQueue::Entry *res = m_queue->GetEntry(timestamp);
  if(res!=NULL)
  {
    res->DecrPending();
    s_debug->print("RACE_SnapshotTrigger::InformInsert(%s,%ld) %d pending",
                   m_description,timestamp,res->GetPending());

    /*
     * If the next pending snapshots don't wait for pending Dumpables
     * Log them too
     * Beginning from beginn
     * If a later snapshot would be ready before a earlier,
     * it would be triggered if its decendant is flushed via this loop
     */
    c=1;
    while(!m_queue->IsEmpty() && m_queue->GetTopPending() == 0)
    {
      s_debug->print("RACE_SnapshotTrigger::InformInsert(%s): triggering %ld (%d in batch)",
                     m_description,m_queue->GetTopTimestamp(),c);
      RACE_SnapshotQueue::Entry* tolog = m_queue->dequeue();
      tolog->Log();
      delete tolog;
      c++;
    }
  }
  else
  {
    RACE_ERROR("RACE_SnapshotTrigger::InformInsert(%s,%ld): not found",
                   m_description,timestamp);
  }
}

