#include"RACE_SnapshotTrigger.h"
#include<iostream>

class CMySnapshotCreator;

//##ModelId=3F16A4870099
class CTestSnapshotEntry : public RACE_SnapshotQueue::Entry
{
public:
    //##ModelId=3F16A48F02D2
  CTestSnapshotEntry(time_t timestamp,int pending,CMySnapshotCreator *sc)
   : RACE_SnapshotQueue::Entry(timestamp,pending),mp_sc(sc) {}
    //##ModelId=3F16A48F02DE
  virtual void Log();
private:
    //##ModelId=3F16A48F02C7
  CMySnapshotCreator *mp_sc;
};

//##ModelId=3EFAEA1102E6
class CMySnapshotCreator : public RACE_SnapshotCreator
{
public:
    //##ModelId=3EFAEA120116
  CMySnapshotCreator(bool totalknown=false)
    :m_pendingtotal(0),m_pendingcurrent(0),m_totalknown(totalknown) {};
    //##ModelId=3EFAEA120118
  virtual int GetPendingTotal();
    //##ModelId=3F16A48B02AE
  virtual int GetPendingForYou();
    //##ModelId=3F16A48B02B0
  virtual RACE_SnapshotQueue::Entry *GenNewEntry(time_t timestamp,int pending);
    //##ModelId=3EFAEA12011F
  void SetPending(int p) { m_pendingtotal=p; m_pendingcurrent=0; }
    //##ModelId=3F16A48B02B6
  void IncrPending(int a) { m_pendingcurrent=a;m_pendingtotal+=a; }
    //##ModelId=3F16A48B02BA
  void SetAndIncrPending(int b,int a)
  {
    m_pendingcurrent=a;
    m_pendingtotal=b+a;
  }
private:
    //##ModelId=3EFAEA120114
  int m_pendingtotal;
    //##ModelId=3F16A48B02A4
  int m_pendingcurrent;
    //##ModelId=3F16A48B02A7
  int m_totalknown;
};


//##ModelId=3EFAEA120118
int CMySnapshotCreator::GetPendingTotal()
{
  return (m_totalknown ? m_pendingtotal : ST_UNKOWN);
}

//##ModelId=3F16A48B02AE
int CMySnapshotCreator::GetPendingForYou()
{
  return (m_totalknown ? ST_UNKOWN : m_pendingcurrent);
}

//##ModelId=3F16A48F02DE
void CTestSnapshotEntry::Log()
{
  std::cerr << "CTestSnapshotEntry::Log() " << GetTimestamp() << std::endl;
  mp_sc->SetPending(0);
}

//##ModelId=3F16A48B02B0
RACE_SnapshotQueue::Entry *CMySnapshotCreator::GenNewEntry(time_t timestamp,int pending)
{
  CTestSnapshotEntry *ret = new CTestSnapshotEntry(timestamp,pending,this);
  return ret;
}

int main(int argc,char **argv)
{
  bool totalknownmode = (argc>1 && strcmp(argv[1],"-t")==0);
  CMySnapshotCreator mysc(totalknownmode);
  RACE_SnapshotTrigger   st(&mysc,(totalknownmode ? "test-totalknown" : "test-currentknown"));

  std::cerr << "Normal Operation" << std::endl;
  mysc.SetAndIncrPending(0,3);
  st.PrepareSnapshot(1);
  st.InformInsert();
  st.InformInsert();
  st.InformInsert();

  std::cerr << (st.IsSnapshotPending() ? "ERROR, a snapshot is pending "
                                       : "OK, no snapshot is pending "
                )
            << std::endl;
  std::cerr << "3 Snapshots " << std::endl;
  mysc.SetAndIncrPending(0,2);
  st.PrepareSnapshot(2);
  std::cerr << (st.IsSnapshotPending() ? "OK, a snapshot is pending "
                         : "ERROR, no snapshot is pending "   )
            << std::endl;
  mysc.IncrPending(2);
  st.PrepareSnapshot(3);
  mysc.IncrPending(2);
  st.PrepareSnapshot(4);
  for(int i=0;i<6;i++)
  {
    st.InformInsert();
  }
  std::cerr << (st.IsSnapshotPending() ? "ERROR, a snapshot is pending "
                                       : "OK, no snapshot is pending "
                )
            << std::endl;

  std::cerr << "Mixed " << std::endl;
  mysc.SetAndIncrPending(0,3);
  st.PrepareSnapshot(5);
  st.InformInsert();
  st.InformInsert();
  mysc.SetAndIncrPending(1,2);
  st.PrepareSnapshot(6);
  st.InformInsert();
  st.InformInsert();
  st.InformInsert();
  std::cerr << (st.IsSnapshotPending() ? "ERROR, a snapshot is pending "
                                       : "OK, no snapshot is pending "
                )
            << std::endl;

  std::cerr << "Nothing pending" << std::endl;
   mysc.SetAndIncrPending(0,0);
  st.PrepareSnapshot(7);

  std::cerr << "3 Snapshots with the same content " << std::endl;
  mysc.SetAndIncrPending(0,3);
  st.PrepareSnapshot(8);
  mysc.IncrPending(0);
  st.PrepareSnapshot(9);
  st.PrepareSnapshot(10);
  st.InformInsert();
  st.InformInsert();
  st.InformInsert();
  std::cerr << (st.IsSnapshotPending() ? "ERROR, a snapshot is pending "
                                       : "OK, no snapshot is pending "
                )
            << std::endl;

  std::cerr << "3 Snapshots with the same content, multiple Prepare " << std::endl;
  mysc.SetAndIncrPending(0,3);
  st.PrepareSnapshot(11);
  mysc.IncrPending(0);
  st.PrepareSnapshot(12);
  st.PrepareSnapshot(13);
  st.PrepareSnapshot(13);
  st.PrepareSnapshot(13);
  st.InformInsert();
  st.InformInsert();
  st.InformInsert();
  std::cerr << (st.IsSnapshotPending() ? "ERROR, a snapshot is pending "
                         : "OK, no snapshot is pending "
  )
     << std::endl;

}

