#include "RACE_Depository.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <utility/errorhandling.h>
const int PART_DEPOSITORY=-1;

static envdebug *debug = new envdebug("Persistence_RACE_Depository");

//##ModelId=3F72AF43004B
RACE_Depository::RACE_Depository():RACE_DepositoryNames(),m_fd(-1),wasrecovered_(false)
{
}

const char *RACE_Depository::GetBasename() const
{
  return GetBaseFilename();
}

void RACE_Depository::MoveAway()
{
  int r;
  output::cond_fail(!IsInitialized(),"RACE_Depository::MoveAway(): Filenames not Initialized");
  r=rename(GetFilename(),GetOldFilename());
  output::cond_fail(r<0,"rename(%s,%s): %s",GetFilename(),GetOldFilename(),
                  strerror(errno));
  index_.SaveOldIndex(GetBaseFilename(),1);
}

void RACE_Depository::RemoveOld()
{
  output::cond_fail(!IsInitialized(),"RACE_Depository::Removeold(): Filenames not Initialized");
  debug->print("Remove old %s ",GetBaseFilename());
  unlink(GetOldFilename());
  unlink(GetUncompleteNewFilename());
  index_.Remove(GetBaseFilename(),1);
  index_.Remove(GetBaseFilename(),2);
}

void RACE_Depository::Remove()
{
   output::cond_fail(!IsInitialized(),"RACE_Depository::Remove(): Filenames not Initialized");
  debug->print("Remove %s ",GetBaseFilename());
  unlink(GetFilename());
  index_.Remove();
}

bool RACE_Depository::SaveUncompleteNew()
{
  int r;
  if(!CheckAndRemoveZeroLengthNew())
  {
  output::warn("RACE_Depository::CheckUncompleteNew: %s uncompleted",GetFilename());
  r=::rename(GetFilename(),GetUncompleteNewFilename());
  output::cond_error(r<0,"RACE_Depository::CheckUncompleteNew:rename(%s,%s): %s",
		       GetFilename(),GetUncompleteNewFilename(),
		       strerror(errno));
  index_.SaveOldIndex(GetBaseFilename(),2);
  return true;
}
  else
  {
    return false;
  }
}

bool RACE_Depository::CheckAndRemoveZeroLengthNew()
{
  struct stat statbuf;
  int r;
  r=::stat(GetFilename(),&statbuf);
  if(r<0)
  {
    if(errno==ENOENT) return true; 
    output::error("RACE_Depository::CheckAndRemoveZeroLengthNew:stat(%s):%s",
		  GetFilename(),
		  strerror(errno));  
    statbuf.st_size=0; // regard als zero length
  }
  if(statbuf.st_size == 0)
  {
    unlink(GetFilename());
    unlink(GetIndexname());
    return true;
  }
  return false;
}

//##ModelId=3F72AF43004D
void RACE_Depository::Open(const char *type_name)
{
  int retry=0;
  wasrecovered_=false;
  initialize(type_name);

  debug->print("RACE_Depository::Open(%s)\n",type_name);
  do
  {
    retry=0;
    m_fd = ::open(GetFilename(),O_RDWR|O_CREAT|O_EXCL,0666);
    if(m_fd<0)
    {
      if(errno==EEXIST)
      {
	if(!ExistsOld())
        {
	  if(CheckAndRemoveZeroLengthNew())
	  {
	    // We have nothing to recover
	    retry=1;
	    wasrecovered_=false;
	    debug->print("RACE_Depository::Open(%s): zero length - ignoring\n",
			 type_name);
	  }
	  else
	  {
	    // The new is saved as old and must be recoved
	    MoveAway();
	    retry=1;
	    wasrecovered_=true;
	    debug->print("RACE_Depository::Open(%s): normal reading of an old Depository\n",
			 type_name);
	    
	}
	  // Good state
	}
	else
	{
	  //We have already an old Depository 
	  bool bad=SaveUncompleteNew();
	  retry=1;
	  wasrecovered_=true;
	  debug->print("RACE_Depository::Open(%s): old %s file exists",type_name,
		       (bad ? "and uncompleted new " : ""));
	  // State may be bad
	}
	/* in file GetOldFilename() is the old state */
      }
      else
      {
	output::fail("Open: %s failed: %s ",GetFilename(),strerror(errno));
      }
    }
  } while(retry>0);
  index_.Open(GetBaseFilename());
}

void RACE_Depository::Open()
{
  RACE_COND_FAIL(m_fd<1,"RACE_Depository::Open(): No file open before ");
}

void RACE_Depository::SwitchPart()
{
  Reopen();
}

//##ModelId=3F72AF43004F
void RACE_Depository::Close()
{
  ::close(m_fd);
  m_fd=-1;
  index_.Close();
}

void RACE_Depository::Reopen()
{
  debug->print("RACE_Depository::Reopen: reopen %s",GetFilename());
  MoveAway();
  Close();
  m_fd = ::open(GetFilename(),O_RDWR|O_CREAT|O_EXCL,0666);
  output::cond_fail(m_fd<0,"Open: %s failed: %s ",GetFilename(),strerror(errno));
  index_.Open(GetBaseFilename());
}

//##ModelId=3F72AF430056
long RACE_Depository::PartNo()
{
  return PART_DEPOSITORY;
}

//##ModelId=3F72AF430051
int RACE_Depository::WriteV(const  struct  iovec  *iov,  int count)
{
  return ::writev(m_fd,iov,count);
}

