#ifndef RACE_LOGFILE_READER_H
#define RACE_LOGFILE_READER_H
#include "RACE_LogfileOffset.h"
#include "RACE_ReadBuffer.h"

namespace RACE
{
  class LogFileReader : public BlockReader
  {
  public:
    LogFileReader(RACE::FullBufferHandlingStrategy *fbhs=0);
    virtual ~LogFileReader();
    const char *GetLogname() { return  logfilename_; }
    const char *GetIndexFilename() { return logfileindexname_; }

    bool Open(const char *logfilename);
    virtual RACE_LogfileOffset GetPosition();

    //##Documentation
    //## after calling lo points to the begin of the Dumpable
    //## and m_pos after the Dumpable
    RACE::DataBlock *Read(bool &fragmented,RACE_LogfileOffset &lo)
      throw (ReadError);
 
    bool IsEndofLog();

    //##Documentation
    //## Positions the logfile at the given offset and
    //## reads and interprets the entry there.
    RACE::DataBlock *ReadAt(
      //##Documentation
      //## An offset in the logfile.
			    RACE_LogfileOffset offset,
			    bool &fragmented,
			    RACE_LogfileOffset &lo
    ) throw (ReadError);
    //##Documentation
    //##  Set the position of the logfile to offset.
    void Seek(RACE_LogfileOffset offset) throw (ReadError);
    void Close() { CloseCurrent(); }

    //##Documentation
    //## Interface BlockReader
    size_t ReadBlock(Octet_t *puffer);
    void CorrectPos(off_t rel);
  protected:
    bool OpenCurrent(const char *filename);
    bool Openpart();
    //##Documentation
    //## after calling lo points to the begin of the Dumpable
    //## and m_pos after the Dumpable
    RACE::DataBlock *ReadCurrent(bool &fragmented,RACE_LogfileOffset& lo) throw (ReadError);
    void CloseCurrent();
    bool SwitchPart();
    enum { LOGTYPE_LOGFILE, LOGTYPE_DEPOSITORY } m_logtype;
  private:
 
    char logfilename_[256];
    char logfileindexname_[256];
    int fd_;
    //##Documentation
    //## Seek() has to maintain the offset (the libc systemcall does not.)
    //## It is set in OpenCurrent().
    off_t m_sizeofcurrentpart;
    bool m_IsEndofLog;
    RACE_LogfileOffset m_pos;
    //##Documentation
    //## if wantedpart > 0 only over this is iterated
    int m_wantedPart; 

    ReadBuffer buffer_;
  };
  const int PART_UNDEFINED=-2;
  const int PART_DEPOSITORY=-1;
};

#endif
