#include"RACE_ReadBuffer.h"
#include<utility/errorhandling.h>

static envdebug *debug=new envdebug("RACE_ReadBuffer");
using namespace RACE;

RACE::FullBufferHandlingStrategy::~FullBufferHandlingStrategy()
{
}

RACE::Index_t RACE::FullBufferHandlingStrategy::notifyBufferIsFull() throw (BufferFull)
{
  throw BufferFull();
  return 0;
}

void RACE::FullBufferHandlingStrategy::notifyPageGetFree(Index_t freepage)
{
}

RACE::ReadBuffer::ReadBuffer(Index_t pagecount,BlockReader *br,FullBufferHandlingStrategy *fbhs):indexmax_(pagecount),blockreader_(br),fullbufferhandling_(fbhs)
{
  Index_t i;
  pdtable_= new Pagedescriptor[indexmax_];
  RACE_MEMORY_FAIL(pdtable_==0,__FILE__,__LINE__,__FUNCTION__);
  puffer_=new Octet_t[(1<<12)*indexmax_];
  RACE_MEMORY_FAIL(puffer_==0,__FILE__,__LINE__,__FUNCTION__);
  for(i=0;i<indexmax_;i++)
  {
    pdtable_[i].pageno=i;
  }
  freepage_=-1;
  offset_=0;
  currentdump_=0;
  if(fullbufferhandling_==0)
  {
    fullbufferhandling_=new FullBufferHandlingStrategy(this);
  }
  else
  {
    fullbufferhandling_->SetReadBuffer(this);
  }
}

RACE::ReadBuffer::~ReadBuffer()
{
}

RACE::Octet_t *RACE::ReadBuffer::index2addr(Index_t i)
{
  RACE_COND_FAIL(i>=indexmax_,"ReadBuffer::index2addr: Indexoverflow");
  return puffer_+(i<<12);
}

RACE::Index_t RACE::ReadBuffer::addr2index(Octet_t *addr)
{
  Index_t ret;
  RACE_COND_FAIL(addr < puffer_,"ReadBuffer::addr2index: Indexoverflow");
  ret = (addr-puffer_)>>12;
  RACE_COND_FAIL(ret>=indexmax_,"ReadBuffer::addr2index: Indexoverflow");
  return ret;
}

RACE::Octet_t *RACE::ReadBuffer::nextpage(bool &directsuccessor) throw (RACE::BufferFull)
{
  Index_t oldpage=freepage_;
  Index_t tries=0;
  freepage_=(freepage_+1)%indexmax_;
  while(pdtable_[freepage_].count>0 && tries<indexmax_) 
  {
    freepage_=(freepage_+1)%indexmax_;
    tries++;
  }
  if(pdtable_[freepage_].count>0)
  {
    freepage_=fullbufferhandling_->notifyBufferIsFull();
  }
  directsuccessor=(freepage_==oldpage+1);
  return index2addr(freepage_);
}

RACE::DataBlock *RACE::ReadBuffer::generate(size_t) throw (NoMoreDynamicMemory)
{
  return 0;
}

void ReadBuffer::seekcorrection(off_t newoffset)
{
  offset_=newoffset;
  rest_=0;
  currentdump_=0;
}

size_t ReadBuffer::readablock(bool &pageisdirectsuccessor,bool expect) throw (BufferFull,ReadError)
{
  size_t ret;
  
  if(!blockuncomplete_)
  {
    ret = blockreader_->ReadBlock(nextpage(pageisdirectsuccessor));
  }
  else
  {
    debug->print("Last Block was uncomplete - EOF\n");
    ret=0;
  }
  if(ret<0 || (expect && ret==0)) 
  {
    RACE_LogfileOffset myoffset = blockreader_->GetPosition();
    myoffset.offset-=rest_;
    throw RACE::ReadError(myoffset,"Unexpected End of logfile");
  }
  blockuncomplete_=(ret!=blockreader_->getBlocksize());
  return ret;
}

DataBlock *ReadBuffer::getDumpable(bool &fragmented,off_t& offset)
{
  output::component c ("ReadBuffer::getDumpable");
  DataBlock *ret=0;
  DataBlock **currentdb=&ret;
  size_t nextfraglen=0;
  fragmented=false;
  bool pageisdirectsuccessor;
  offset=offset_;
  blockuncomplete_=false;
  try
  {
    if(currentDumpable()==NULL || rest_==0)
    {
      rest_=readablock(pageisdirectsuccessor);
      debug->print("fresh rest_=%d",rest_);      
      if(rest_==0) return 0;
      
      currentdump_=currentpage();
    }
    if(rest_<sizeof(Layout::GeneralHeader))
    {
      debug->print("rest_=%d<sizeof(Layout::GeneralHeader)",rest_);
      Octet_t *frag=currentdump_;
      size_t n=readablock(pageisdirectsuccessor,true);
      debug->print("read next : n=%d",n);
      if(pageisdirectsuccessor)
      {
	// No problem
	rest_+=n;
      }
      else
      {
	// Correct it - we have to copy and one seek
	// (On other possibility would be read less than intended) 
	Octet_t *cur = currentpage();
	memmove(cur+rest_,cur,n-rest_);
	memcpy(cur,frag,rest_);
	blockreader_->CorrectPos(-(off_t)rest_);
	rest_=n;
	currentdump_=cur;
      }
    }
    debug->print("Dumpable type %ld length %ld",
		 currentDumpable()->type,
		 currentDumpable()->len);
    
    nextfraglen=(size_t)(currentDumpable()->len);
    offset_+=(size_t)(currentDumpable()->len);

    while(nextfraglen>rest_)
    {
      size_t n=readablock(pageisdirectsuccessor,true);
      debug->print("bigread next : n=%d rest_=%d len=%d",n,rest_,nextfraglen);
      if(pageisdirectsuccessor)
      {
	// No Problem while buffer contiguous
	rest_+=n;
      }
      else
      {
	// uncontiguous memory
	// Solutions: fragment instead copying
	nextfraglen-=rest_;
	debug->print("Uncontigous -> Fragment len=%d nextfraglen=%d",
		     rest_,nextfraglen);
	*currentdb=create(this,(char *)currentdump_,rest_,true);
	mark(*currentdb);
	currentdb=&((*currentdb)->next);
	rest_=n;
	currentdump_=currentpage();
	fragmented=true;
      }
    }
    *currentdb=create(this,(char *)currentdump_,nextfraglen,true);
    mark(*currentdb);
    rest_-=nextfraglen;
    debug->print("len=%d rest_=%d",nextfraglen,rest_);
    debug->print("generate Dumpable on page %ld count=%ld",freepage_,pdtable_[freepage_].count);
    currentdump_+=nextfraglen;
    return ret;
  }
  catch(BufferFull& ex)
  {
    dumppagetable();
    RACE_FATAL_ERROR("Buffer Full");
    return 0;
  }
}

void RACE::ReadBuffer::informduplicate(DataBlock *db)
{
  mark(db);
}

void RACE::ReadBuffer::informrelease(DataBlock *db)
{
  unmark(db);
}

void RACE::ReadBuffer::recycle(DataBlock *schrott)
{
  debug->print("RACE::ReadBuffer::recycle datablock ");
  unmark(schrott);
  destroy(schrott);
}

void RACE::ReadBuffer::mark(DataBlock *b)
{
  if(debug->level()>1) dumppagetable();
  Index_t start=addr2index((Octet_t *)(b->begin()));
  Index_t end  =addr2index((Octet_t *)(b->begin())+b->size()-1); 
  debug->print("mark pages %ld .. %ld",start,end);
  RACE_GUARD(ACE_Thread_Mutex,guard,mutex_,__FILE__,__LINE__,__FUNCTION__);
  for(Index_t i=start;i<=end;i++)
  {
    pdtable_[i].count++;
  }
  if(debug->level()>1) dumppagetable();
}

void RACE::ReadBuffer::unmark(DataBlock *b)
{
  if(debug->level()>1) dumppagetable();
  Index_t start=addr2index((Octet_t *)(b->begin()));
  Index_t end  =addr2index((Octet_t *)(b->begin())+b->size()-1); 
  debug->print("unmark pages %ld .. %ld",start,end);
  RACE_GUARD(ACE_Thread_Mutex,guard,mutex_,__FILE__,__LINE__,__FUNCTION__);
  for(Index_t i=start;i<=end;i++)
  {
    pdtable_[i].count--;
    debug->print("unmark(%ld):count=%ld",i,pdtable_[i].count);
    if(pdtable_[i].count==0)
    {
      fullbufferhandling_->notifyPageGetFree(i);
    }
  }
  if(debug->level()>1) dumppagetable();
}

void RACE::ReadBuffer::dumppagetable()
{
  for(Index_t i=0;i<indexmax_;i++)
  {
    RACE_DEBUG("%ld\t%ld",pdtable_[i].pageno,pdtable_[i].count);
  }
}

#ifdef RACE_READER_TEST
void ReadBuffer::addrindextest()
{
  for(Index_t i=0;i<indexmax_;i++)
  {
    Octet_t *a = index2addr(i);
    Index_t i2= addr2index(a);
    RACE_COND_FAIL(i!=i2,"i!=i2");
    Index_t i3= addr2index(a+10);
    RACE_COND_FAIL(i!=i3,"i!=i3");
  }
  Octet_t *a = index2addr(1);
  Index_t ip1 = addr2index(a+4096);
  RACE_COND_FAIL(ip1!=2,"Expected ip1==2");
}


int main()
{
  RACE::ReadBuffer *r=new RACE::ReadBuffer(8);
  r->addrindextest();
}
#endif


