#ifndef RACE_READBUFFER_H
#define RACE_READBUFFER_H
#include<ace/Synch_T.h>
#include<RACE_DataBlock.h>
#include<RACE_Layout_GeneralHeader.h>
#include<RACE_ReadError.h>

namespace RACE
{
  typedef unsigned char Octet_t;
  typedef long Index_t;
  typedef unsigned long Count_t;

  class BlockReader
  {
  public:
    BlockReader(size_t bs=getpagesize()):blocksize_(bs) {}
    virtual RACE_LogfileOffset GetPosition()=0;
    virtual size_t ReadBlock(Octet_t *puffer)=0;
    virtual void CorrectPos(off_t rel)=0;
    size_t getBlocksize() { return blocksize_; }
  protected:
    size_t blocksize_;
  };

  class ReadBuffer;

  //##Documentation
  //## A Exception Class
  class BufferFull {};

  //##Documentation
  //## controls the behavior when the ReadBuffer is full
  //## Note: depended on the realiation, more than one thread may call
  //## different Callbacks
  class FullBufferHandlingStrategy
  {
  public:
    FullBufferHandlingStrategy(ReadBuffer *rb=0):readbuffer_(rb) {}
    virtual ~FullBufferHandlingStrategy();
    void SetReadBuffer(ReadBuffer *rb) { readbuffer_=rb; }
    //##Documentation
    //## This is called from ReadBuffer::nextpage()
    //## When the situation is handeld, it returns an Index 
    //## of the freepage
    virtual Index_t notifyBufferIsFull() throw (BufferFull);
    virtual void notifyPageGetFree(Index_t freepage);
  protected:
    ReadBuffer *readbuffer_;
  };

  class ReadBuffer : public DataBlockFactory
  {
  public:
    struct Pagedescriptor
    {
      Index_t pageno;
      Count_t count;
      Pagedescriptor():pageno(0),count(0) {}
    };
    ReadBuffer(Index_t pagecount,BlockReader *br,FullBufferHandlingStrategy *fbhs=0);

    virtual ~ReadBuffer();
    virtual DataBlock *generate(size_t) throw (NoMoreDynamicMemory);
    //##Documentation
    //## MAY CALLED from an other Thread
    virtual void recycle(DataBlock *schrott);
    //##Documentation
    //## We need to count the accessors on a BufferPage
    //## MAY CALLED from an other Thread
    virtual void informduplicate(DataBlock *db);
    //##Documentation
    //## We need to count the accessors on a BufferPage
    //## MAY CALLED from an other Thread
    virtual void informrelease(DataBlock *db);
    //##Documentation
    //## if we have to wrap or to jump other a unfree block
    //## we chain the DataBlocks and set fragmented==true;
    //## The offset is the local offset in the part
    DataBlock *getDumpable(bool &fragmented,off_t& offset);

    //##Documentation
    //## if we change the position in the Logfile, we have to begin
    //## on a new page and correct the offset.
    void seekcorrection(off_t newoffset);
    off_t getoffset() { return offset_; }
    void dumppagetable();
  protected:
    Octet_t *currentpage() { return index2addr(freepage_); }
    Octet_t *nextpage(bool &directsuccessor) throw (BufferFull);
    Layout::GeneralHeader *currentDumpable()
    {
      return (Layout::GeneralHeader *)currentdump_; 
    }
    //##Documentation
    //## Mark all pages of the datablock
    //## Encapsule the concurreent used data
    void mark(DataBlock *b);
    //##Documentation
    //## Mark all pages of the datablock
    //## Encapsule the concurreent used data
    void unmark(DataBlock *b);
  protected:
    Octet_t *index2addr(Index_t i);
    Index_t addr2index(Octet_t *addr);
    //##Documentation
    //## Helperfunktion, does errorhandling, sezts blockuncomplete_
    size_t readablock(bool &pageisdirectsuccessor,bool ex=false) throw (BufferFull,ReadError);
  private:
    Index_t indexmax_;
    BlockReader *blockreader_;
    FullBufferHandlingStrategy *fullbufferhandling_;
    Octet_t *puffer_;
    Pagedescriptor *pdtable_;
    Index_t freepage_;
    Octet_t *currentdump_;
    off_t offset_;
    size_t rest_;
    bool blockuncomplete_;
    ACE_Thread_Mutex mutex_;
  };
};
#endif
