/*
 * Test fpr Logger and Reader: Writing Dumpables with general headers
 */
#include<iostream>
#include"RACE_Logger.h"
#include"RACE_Logfile.h"
#include"RACE_Layout_GeneralHeader.h"

class CTestDumpable : public RACE_Dumpable
{
  char pat_;
  CTestDumpable(char pat):pat_(pat) {} // For Cloning
  RACE::Layout::GeneralHeader *header;
public:
  CTestDumpable(size_t s,char pat):RACE_Dumpable(s),pat_(pat)
  {
    header = (RACE::Layout::GeneralHeader *)wr_ptr(sizeof(RACE::Layout::GeneralHeader));
    size_t r = freespace();
    header->len=s;
    header->type=s;
    memset(wr_ptr(r),pat,r);
  }
  virtual RACE_Message_Block *CloneManagementData() const;
  virtual void InformOffset(RACE_LogfileOffset offset);
  void expand();
};

void CTestDumpable::expand()
{
  RACE_Message_Block::expand();
  header->len+=size();
}

void CTestDumpable::InformOffset(RACE_LogfileOffset offset)
{
  std::cerr << "Written Pattern '" << pat_ << "' with length=" << total_length()
       << "\tat offset " << offset.part << ':' << offset.offset << std::endl;
}

RACE_Dumpable *genentry(size_t s,char pat)
{
  return new CTestDumpable(s,pat);
}

RACE_Dumpable *genfragmented(size_t s,int count,char pat1,char pat2)
{
  int i;
  CTestDumpable *f = new CTestDumpable(s,pat1);
  for(i=0;i<count-1;i++)
  {
    f->expand();
    memset(f->wr_ptr(s),pat2,s);
  }
  return f;
}

RACE_Message_Block *CTestDumpable::CloneManagementData() const
{
  CTestDumpable* ret = new CTestDumpable(pat_);
  return ret;
}

void simmulator(RACE_Logger *log)
{
  int g=0;
  for(int n=0;n<100 ;n++)
  {
    RACE_Dumpable *d;
    // Allocate a new message.
    if(g<6)
    {
      size_t s= (1<<(g+8));
      char pat='A'+(n%26);
      d = genentry(s,pat);
      std::cout << "Sender: iteration " << n << " g= " << g
           << " size=" << s << " Pattern " << pat << std::endl;
    }
    else
    {
      size_t s=4096;
      char pat1='A'+(n%26);
      char pat2='a'+(n%26);
      d = genfragmented(s,g-4,pat1,pat2);
      std::cout << "Sender: iteration " << n << " g= " << g
           << " Fragmented: " << g-4 << "* size=" << s << " Pattern "
           << pat1 << "/" << pat2 << std::endl;
    }
    log->Log(d);
    g=(g+1)%10;
  }
  log->Cancel();
}

int main(int,char**)
{
  RACE_Logfile logfile;
  logfile.Open("loggertestgh");
  RACE_LoggerStrategy loggerstrat("loggertest");
  RACE_Logger *logger=new RACE_Logger(&logfile,&loggerstrat);
  logger->activate();
  simmulator(logger);
  logger->wait();
  logger->Release();
  
  std::cerr << "done" << std::endl;
  return 0;
}

