#ifndef BITMAP_H
#define BITMAP_H
#include<utility/errorhandling.h>
#include<iostream>
using namespace std;


typedef unsigned long index_t;

/**
 * @brief persistent Bitmap-Structure
 * 
 * Man kann bits setzen und l&ouml;schen und kann das erste freie bit suchen
 */
class BitMap
{
  size_t len;          // anzahl
  unsigned long f[1];  // 32 bit 
 public: // 32=1<<5; 31=(1<<5)-1;
  void init(int plen) { len=plen>>5; for(index_t i=0;i<len;i++) f[i]=0; }
  size_t size() { return sizeof(size_t)+len*sizeof(unsigned long); }
  size_t length() { return len; }
  bool is_set(index_t i)
        { RACE_COND_FAIL((i>>5)>=len," BitMap::Index too big");
	  return f[i>>5]&(1<<(i&31));
	}
  void set(index_t i,bool w)
        { RACE_COND_FAIL((i>>5)>=len," BitMap::Index too big");
	  unsigned long m = (1<<(i&31));
	  if(w) { f[i>>5] |= m; } else { f[i>>5] &= ~m; }
	}

  /// finds the first unset bit
  index_t find0()
    {
      index_t i=0;
      unsigned short j;
      unsigned short o=0;
      while(i < len && f[i] == ~0u) i++;
      o=0;
      while(((0xfu << o)&f[i])==(0xfu<<o) && o < 32u) o+=4;
      for(j=o;j<o+4;j++)
      {
	if(!((1<<j)&f[i])) return((i<<5) + j);
      }
      RACE_COND_FAIL(0,"%s:%d: should not be reached",__FILE__,__LINE__);
      return (~0);
    }
};

inline void druckebelegtliste(ostream &str,BitMap &bm)
{
  for(index_t i=0;i<bm.length();i++)
  {
    if(bm.is_set(i)) str << (i>0 ? " ," : "") << i;
  }
  str << endl;
};
#endif




