#include"LSeg.h"
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
#include<string.h>
#include"LStools.h"

#ifndef _POSIX_MAPPED_FILES
#error "Missing _POSIX_MAPPED_FILES : Implementation only for standard os ! "
#endif
static envdebug debug("RAMS_HVSEG_LSeg");

void LSeg::open(const char *name) throw (Systemerror)
{
  struct stat sbuf;
  int ldnoofpages=4;
  int noofpages=1<<ldnoofpages;
  int r;
  size_t soll=1<<(ldnoofpages+12);
  size_t ist;
  if(pagedescriptor_!=NULL) delete pagedescriptor_;
  pagedescriptor_ = new LSPagedescriptor[noofpages];
  RACE_MEMORY_FAIL(pagedescriptor_==NULL,__FILE__,__LINE__,__FUNCTION__);

  int namelen=strlen(name);
  char *filename=(char *)alloca(namelen+4);
  snprintf(filename,namelen+4,"%s.lseg",name);
  fd_=::open(filename,O_CREAT|O_RDWR,0600);
  if(fd_<0) throw Systemerror(errno,"open"," open(%s): %m",filename);
  r=fstat(fd_,&sbuf);
  if(r<0)  throw Systemerror(errno,"stat"," stat(%s): %m", filename);
  // How big should our segment be ? 
  if(sbuf.st_size == 0)
  {
    debug.print("Create a initial segment of size %d",soll);
    createzeroedfile(filename,fd_,noofpages);
    ist=noofpages<<12;
  }
  else
  {
    ist=sbuf.st_size;
  }
  char *seg=(char *)::mmap(NULL,ist,PROT_READ,MAP_SHARED,fd_,0);
  if(seg==(char *)MAP_FAILED) throw Systemerror(errno,"mmap"," mmap(%s): %m",filename);
  debug.print("LSeg:open: map base=%p,length=%d",seg,ist);
  set(seg,ist);

  // erst nach dem set koennen die Pagesegmente richtig initialisiert weren
  for(RAMS::PageNo_t i=0;i<noofpages;i++)
  {
    pagedescriptor_[i].init(this,i);
  }
  
  RACE_Logfile *logfile = new RACE_Logfile();
  RACE_MEMORY_FAIL(logfile==NULL,__FILE__,__LINE__,__FUNCTION__);
  logfile->Open(name);
  RACE_LoggerStrategy *loggerstrat = new RACE_LoggerStrategy("lsseglogger");
  logger_ = new RACE_Logger(logfile,loggerstrat);
  RACE_MEMORY_FAIL(logger_==NULL,__FILE__,__LINE__,__FUNCTION__);
  logger_->activate();
}

void LSeg::checkpoint()
{
  debug.print("Lseg::checkpoint");
  for(RAMS::PageNo_t i=0;i<maxpageno();i++)
  {
    debug.print("%ld : %s",i,(pagedescriptor_[i].isdirty() ? "dirty" : "clean"));
    pagedescriptor_[i].logchangesifneeded();
  }
  tan_++;
}

void LSeg::rollback()
{
    debug.print("Lseg::rollback");
  for(RAMS::PageNo_t i=0;i<maxpageno();i++)
  {
    debug.print("%ld : %s",i,(pagedescriptor_[i].isdirty() ? "dirty" : "clean"));
    pagedescriptor_[i].rollbackifneeded();
  }
}

void LSeg::close()
{
  if(logger_)
  {
    logger_->Cancel();
    logger_->wait();
    logger_->Release();
    logger_=0;
  }
  if(fd_>-1)
  {
    ::close(fd_);
    fd_=-1;
  }
}

LSeg::~LSeg()
{
  close();
}

void LSeg::handle(RAMS::PageNo_t pageno,void *)
{
  debug.print("Lseg::handle: Pageno %ld",pageno);
  pagedescriptor_[pageno].touch();
}
