#include<utility/errorhandling.h>
#include<string.h>
#include<unistd.h>
#include"LSpagedescriptor.h"
#include"LSeg.h"
#include"LSFullPageDumpable.h"
#include"LSDiffPageDumpable.h"

static envdebug debug("RAMS_HVSEG_LSeg");
RACE::StupidMallocDataBlockFactory smdbf;

void LSPagedescriptor::init(LSeg *lseg,RAMS::PageNo_t i)
{
  lseg_=lseg; index_=i;
  pageseg_.set((char *)(lseg->pageno2addr(i)),PAGESIZE);
}

void LSPagedescriptor::touch()
{
  if(!dirty_)
  {
    if(pageheader_==NULL)
    {
      pageheader_=smdbf.generate(sizeof(LSlayout::logpageheader));
      pageheader_->grow(sizeof(LSlayout::logpageheader));
      pageheader()->len=sizeof(LSlayout::logpageheader);
      pageheader()->type=LSlayout::full;
      pageheader()->pageno=index_;
      pageheader()->timestamp=time(NULL);
      pageheader()->last.part=0;
      pageheader()->last.offset=0;
      pageheader()->credentials.uid=getuid();
      pageheader()->credentials.gid=getgid();
      pageheader()->offset_in_page=0;
      pageheader()->len_of_page=4096;
    }
    pageheader()->tan = lseg_->tan();
    if(beforeimage_==NULL)
    {
      beforeimage_=smdbf.generate(4096);
      beforeimage_->grow(4096);
    }
    debug.print("touch: Copy page %ld",index_);
    memcpy((char *)(beforeimage_->begin()),lseg_->pageno2addr(index_),4096);
  }
  dirty_=true;
}

void LSPagedescriptor::logchanges()
{
  debug.print("LSPagedescriptor::logchanges: page %ld",index_);
  LSDumpable *lsd=NULL;
  count_++;
  // At the first time (and for snapshots) we want to log the old-image
  // The Changes have to be logged in all cases for reconstructing
  if(count_%10==0 ) 
  {
    debug.print("Log whole page of page %ld ",index_);
    lsd = new LSFullPageDumpable(pageheader_,beforeimage_->duplicate());
    RACE_MEMORY_FAIL(lsd==NULL,__FILE__,__LINE__,__FUNCTION__);
    lseg_->logger_->Log(lsd);
  }
  debug.print("Log diff of Page %ld",index_);
  lsd = new LSDiffPageDumpable(pageheader_,&pageseg_,beforeimage_);
  RACE_MEMORY_FAIL(lsd==NULL,__FILE__,__LINE__,__FUNCTION__);
  beforeimage_->release();
  beforeimage_=NULL;

  lseg_->logger_->Log(lsd);
  lseg_->protect(index_);
  dirty_=false;
}

void LSPagedescriptor::rollback()
{
    debug.print("LSPagedescriptor::rollback: page %ld",index_);
    RACE_COND_FAIL(beforeimage_==NULL,"No beforimage for page %ld",index_);
    memcpy(lseg_->pageno2addr(index_),(char *)(beforeimage_->begin()),4096);
    lseg_->protect(index_);
    dirty_=false;
}
