#include<sys/mman.h>
#include<fcntl.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/errno.h>
#include<stdlib.h>
#include"shadowsegment.h"
#include<iostream>
#include<assert.h>


static envdebug debug("RAMS_HVSEG_shadowsegment");

SegmentVersion *SegmentVersion::init(char *svpuffer,int len)
{
  SegmentVersion *ret = (SegmentVersion *)svpuffer;
  ret->count_ = 0;
  Pageentry bpe;
  bpe.shadowbit=0;
  bpe.pageno=~0;

  for(unsigned int i=0;i< ret->maxcount();i++)
  {
    (*ret)[i]=bpe;
  }

  return ret;
}

int SegmentVersion::clearallshadowbits()
{
  int anz=0;
  for(unsigned int i=0;i<count();i++)
  {
    if((*this)[i].shadowbit)
    {
      (*this)[i].shadowbit=0;
      anz++;
    }
  }
  return anz;
}

void SegmentVersion::print(std::ostream& str)
{
  for(index_t i=0;i<count();i++)
  {
    str << i << ( entry[i].shadowbit ? "*" : " " );
    if(entry[i].pageno == UNDEFINED_PAGE) str << "UNDEF";
    else str << entry[i].pageno;
    str << std::endl;
  }
  str << std::endl;
}

/*------------ Shadow Master --------------*/
ShadowSegMaster::ShadowSegMaster()
    :tws(),rseg(&tws),
    mc(&rseg,((MasterLayout *)rseg.begin())->mc),
    m0(&rseg,((MasterLayout *)rseg.begin())->m0),
    m1(&rseg,((MasterLayout *)rseg.begin())->m1)
{
}

int ShadowSegMaster::open(char *name)
{
  debug.print("ShadowSegMaster::open(%s)\n",name);
  if(rseg.open(name)==1)
  { 
    debug.print("\tInitialisation ShadowSegMaster\n");
    m0.setoff(sizeof(MasterLayout));
    m0.ptr()->init(MAXPAGES);
    m1.setoff(sizeof(MasterLayout)+m0.ptr()->size());
    m1.ptr()->init(MAXPAGES);
    mc=m0;
    setstatus(0);
    rseg.write_page(0);
    return 1;
  }
  else
  {
    debug.print("\tread ShadowSegMaster\n");
    rseg.read_page_reliable(0);
    return 0;
  }
}

void ShadowSegMaster::save()
{
  debug.print("Saving Segmaster\n");
  rseg.save();
}

void ShadowSegMaster::copyandswitchm()
{
  RAMS::RPointer<BitMap> mn;
  debug.print("copyandswitchm()\n");
  mn = (mc==m0 ? m1 : m0);
  memcpy(mn.ptr(),mc.ptr(),mc.ptr()->size());
  mc=mn;
}

void ShadowSegMaster::druckebelegtliste()
{
  std::cerr << "Belegt in CurrentM:";
  ::druckebelegtliste(std::cerr,CurrentM());
}

/*----------------- Shadowsegment -----------------*/
void ShadowSegment::init()
{
  debug.print("Initialisation Segments\n");
  memset(svpuffer,'\0',2*PAGESIZE);
  vw[0]=SegmentVersion::init(svpuffer,PAGESIZE);
  vw[1]=SegmentVersion::init(svpuffer+PAGESIZE,PAGESIZE);
  savesegversion(0,0);
  savesegversion(1,1);
}

void ShadowSegment::clearshadowbitsandfree()
{
  debug.print("ShadowSegment::clearshadowbitsandfree:\n");
  debug.print("Vwork.count = %d",Vwork().count());
  for(index_t i=0;i<Vwork().count();i++)
  {
    debug.print("%ld %s %ld ",i,(Vwork()[i].shadowbit ? "*" : " "),Vwork()[i].pageno);
    if(Vwork()[i].shadowbit)
    {
      if(Vold()[i].pageno != UNDEFINED_PAGE)
      {
	debug.print("\tFree old Page %ld ",Vold()[i].pageno);
	ssm.CurrentM().set(Vold()[i].pageno,false);
      }
      else
      {
	debug.print("\tWas a new page ");
      }
      Vwork()[i].shadowbit = 0;
    }
  }
}

void ShadowSegment::open(const char *name) throw (Systemerror)
{
  char mastername[64];
  char segname[64];
  char versionname[64];
  int r;

  debug.print("ShadowSegment::open(%s)\n",name);
  snprintf(mastername,64,"%s.master",name);
  snprintf(versionname,64,"%s.version",name);
  snprintf(segname,64,"%s.seg",name);

  vfd = ::open(versionname,O_RDWR|O_CREAT,0660);
  if(vfd < 0)
  {
    throw Systemerror(errno,"open","ShadowSegment::init:open(%s):%m",versionname);
  }
  if(ssm.open(mastername)==1)
  {
    init();
  }
  else
  {
    if(ssm.getstatus()==1)
    {
       recover();
    }
    readsegversions();
    if(debug.is_on()){ ssm.druckebelegtliste(); }
  }
  if(debug.is_on())
  {
    std::cerr << "V0:" << endl; Vwork().print(std::cerr);
    std::cerr << "V1:" << endl; Vold().print(std::cerr);
  }
  /** Gebe alte Seiten Frei und lsche Schadowbits - Abweichung **/ 
  clearshadowbitsandfree(); 
  if(debug.is_on())
  {
    std::cerr << "Nach clearshadowbitsandfree():" << endl;
    std::cerr << "V0:" << endl; Vwork().print(std::cerr);
    std::cerr << "V1:" << endl; Vold().print(std::cerr);
  }
  /** Kopiere Vi0 nach Vi1 auf der Platte */
  savesegversion(0,1);
  if(debug.is_on())
  {
    std::cerr << "Nach savesegversion(0,1):" << endl;
    std::cerr << "V0:" << endl; Vwork().print(std::cerr);
    std::cerr << "V1:" << endl; Vold().print(std::cerr);
  }
  /** Setze Status=1 */
  ssm.setstatus(1);
  /** schreibe Master auf Platte **/
  ssm.save();
  if(debug.is_on()) { ssm.druckebelegtliste(); }
  /** loesche alle Shadowbits in vi0 */
  /// Abweichung vw[0]->clearallshadowbits();

  sfd = ::open(segname,O_RDWR|O_CREAT,0660);
  if(sfd < 0) 
  {
    throw Systemerror(errno,"open","ShadowSegment::init:open(%s):%m",segname);
  }

  char *alloc=NULL;
  char *first=NULL;
  size_t lang=0;
  debug.print("Mappe %d Seiten ...",Vwork().count());
 
  if(Vwork().count()==0)
  {
    // grow
    int anzseiten=10;
    debug.print("Neues Segment, lege %d  Seiten an ...",anzseiten);
    Vwork().grow(anzseiten);
    long lang = anzseiten*getpagesize();
    int f0 = ::open("/dev/zero",O_RDONLY);
    if(r==-1) throw Systemerror(errno,"open","ShadowSegment::open(/dev/zero):%m");
    char *b=(char *)mmap(NULL,lang,PROT_READ,MAP_PRIVATE,f0,0);
    if(b==MAP_FAILED) 
    {
      throw Systemerror(errno,"mmap","ShadowSegment::init:mmap(%s):%m",segname);
    } 
    set(b,lang);
    debug.print("Segment gemappt: %p - %p",(void *)begin(),(void *)end());
    return; // 1;
  }
  {
    /* Reservierung, denn wer weiss wo die Seiten plaziert werden */
    int anzseiten=10;
    long lang = anzseiten*getpagesize();
    int f0 = ::open("/dev/zero",O_RDONLY);
    first=alloc=(char *)mmap(NULL,lang,PROT_READ,MAP_PRIVATE,f0,0);
    if(alloc==MAP_FAILED) 
    {
      throw Systemerror(errno,"mmap","ShadowSegment::init:mmap(%s):%m",segname);
    } 
    munmap(alloc,lang);
    set(alloc,lang);
    debug.print("Segment reserviert: %p - %p",(void *)begin(),(void *)end());
  }
  for(index_t i=0;i<Vwork().count();i++)
  {
    if(Vwork()[i].pageno != UNDEFINED_PAGE)
    {
      off_t off = (Vwork()[i].pageno  << 12);
      debug.print("Seite %ld  Block %ld %p <- %ld",i,Vwork()[i].pageno,
		  (void *)alloc,off);
      char *b=(char *)mmap(alloc,PAGESIZE,PROT_READ,MAP_SHARED,sfd,off);
      if(b==MAP_FAILED) 
      {
	throw Systemerror(errno,"mmap","ShadowSegment::init:mmap(%s):%m",segname);
      }
      if(!alloc)
      {
	alloc = b;
	first=b;
	debug.print("Erstallocation bei %p",(void *)alloc);
      }      
      else if(b != alloc)
      {
	throw Systemerror(0,"mmap","Shadowsegment::open: gewuenschte Speicheradresse nicht erhalten ");
      }
    }
    alloc+=PAGESIZE;
    lang +=PAGESIZE;
  }
  set(first,lang);
  debug.print("Segment gemappt: %p - %p",(void *)begin(),(void *)end());
  // return; // 0;
}

void ShadowSegment::recover()
{
  debug.print("Recover from Crash\n");
  readsegversions();
  savesegversion(1,0);
  if(debug.is_on())
  {
    std::cerr << "V0 und V1:" << endl;Vwork().print(std::cerr);
    ssm.druckebelegtliste();
  }
  debug.print("\tSetze status=0\n");
  ssm.setstatus(0);
  ssm.save();
}  
  

void ShadowSegment::handle(RAMS::PageNo_t pageno,void *addr)
{
  index_t dp;
  off_t dpo;
  debug.print("ShadowSegment::handle page(%ld)",pageno);
  char *page=(char *)pageno2addr(pageno);
  dp= ssm.CurrentM().find0();
  ssm.CurrentM().set(dp,true);

  
  if(debug.is_on())
  {
    std::cerr << "alte Seite: ";
    if(Vwork()[pageno].pageno == UNDEFINED_PAGE )
      std::cerr << "undefined";
    else 
      std::cerr <<  Vwork()[pageno].pageno;
  }
  
  dpo= index2offset(dp);
  if(debug.is_on())
  {
    std::cerr << " neue Seite: " << dp << " (Offset=" << dpo << ")" << endl;
  }

  int r=lseek(sfd,dpo,SEEK_SET);
  if(r<0) { throw Systemerror(errno,"lseek","handle pagefault: lseek: %m"); }
  r=write(sfd,page,PAGESIZE);
  if(r<0) { throw Systemerror(errno,"write","handle pagefault: write: %m"); }
  r=munmap(page,PAGESIZE);
  if(r<0) { throw Systemerror(errno,"munmap","munmap %m"); }
  /// schreibbar, denn in zukunft wird gleich der Block modifiziert
  void *newp = mmap(page,PAGESIZE,PROT_READ|PROT_WRITE,MAP_FIXED|MAP_SHARED,sfd,dpo);
  if(newp==(void *)(-1)) { throw Systemerror(errno,"mmap","mmap %m"); }

  Vwork()[pageno].pageno=dp;
  Vwork()[pageno].shadowbit=1;
}

void ShadowSegment::readsegversions() throw(Systemerror)
{
  debug.print("readsegversions()\n");
  char *p=svpuffer;
  int rest=2*PAGESIZE;

  int nr=0;
  while(rest>0 && (nr= ::read(vfd,p,rest))>0)
  {
    p+=nr;
    rest-=nr;
  }
  if(nr < 0) throw Systemerror(errno,"read",
			       "ShadowSegment::readsegversions: read:%m");
  vw[0]=(SegmentVersion *)svpuffer;
  vw[1]=(SegmentVersion *)(svpuffer+PAGESIZE);
}

void ShadowSegment::savesegversion(int nr,int snr) throw(Systemerror)
{
  debug.print("savesegversion v%d auf Platte v%d\n",nr,snr);
  assert(0<=nr && nr<2);
  assert(0<=snr && snr<2);
  off_t fo = snr*PAGESIZE;
  int r=lseek(vfd,fo,SEEK_SET);
  if(r<0) throw Systemerror(errno,"lseek","savesegversion(%d):lseek:%m",nr);
  r=write(vfd,(char *)vw[nr],PAGESIZE);
  if(r<0) throw Systemerror(errno,"write","savesegversion(%d):write:%m",nr);
  if(nr!=snr)
  {
    memcpy((char *)vw[snr],(char *)vw[nr],vw[nr]->size());
  }
}

void ShadowSegment::save()
{
  debug.print("------- Save Segment --------\n");
  if(debug.is_on()) { ssm.druckebelegtliste(); }
  ssm.copyandswitchm();
 
  if(debug.is_on()) ssm.druckebelegtliste();
  
  for(index_t i=0;i<Vwork().count() && i<<12 < size();i++)
  {
    off_t o  = index2offset(i);;
    off_t fo = index2offset(Vwork()[i].pageno);
   
    if(Vwork()[i].pageno == UNDEFINED_PAGE)
    {
      debug.print("Seite %ld unbenutzt",i);
    }
    else if(Vwork()[i].shadowbit)
    { 
      debug.print("Speichere Seite %ld (%ld) -> %ld (%ld)",
		  i,o,Vwork()[i].pageno,fo);
      void *addr = (char *)begin()+o; 
      int r=msync(addr,PAGESIZE,MS_SYNC);
      if(r<0) { throw Systemerror(errno,"msync","msync %m"); }
      r=mprotect(addr,PAGESIZE,PROT_READ);
      if(r<0) { throw Systemerror(errno,"mprotect","mprotect %m"); }
#if 0 // With Write
      int r=lseek(sfd,fo,SEEK_SET);
      if(r<0) { throw Systemerror(errno,"lseek","lseek %m"); }
      r=write(sfd,akt.ptr(),PAGESIZE);
      if(r<0) { throw Systemerror(errno,"write","write %m"); }
#endif
    }
    else
    {
      debug.print("Seite %ld unveraendert ",i);
    }
  }
  fsync(sfd);
  savesegversion(0,0);
  ssm.setstatus(0);
  if(debug.is_on()) { ssm.druckebelegtliste(); }
  ssm.save();
  debug.print("End von Save \n");
  if(debug.is_on())
  {
    std::cerr << "V0:" << endl; Vwork().print(std::cerr);
    std::cerr << "V1:" << endl; Vold().print(std::cerr);
  }
}


void ShadowSegment::reopen()
{
  clearshadowbitsandfree();
  savesegversion(0,1);
  ssm.setstatus(1);  
  ssm.save();
}

void ShadowSegment::rollback()
{
  debug.print("rollback\n");
  for(index_t i=0;i<Vwork().count() && i<<12 < size();i++)
  {
    if(Vwork()[i].shadowbit)
    {
      debug.print("%ld instead %ld -> %ld ",i,Vwork()[i].pageno,Vold()[i].pageno);
       
      // Seite mu wieder Richtig gemappt werden
      off_t o=i<<12;
      off_t old_off=Vold()[i].pageno <<12;
      void *addr = (char *)begin()+o;
      if(munmap(addr,PAGESIZE)<0)
      {
	throw Systemerror(errno,"munmap","munmap %m");
      }
      ssm.CurrentM().set(Vwork()[i].pageno,false);
      char *b=(char *)mmap(addr,PAGESIZE,PROT_READ,MAP_SHARED,sfd,old_off);
      assert(b!=MAP_FAILED && (void *)b==addr);
      Vwork()[i] = Vold()[i];
    }
  }
}









