#ifndef COLLECTOR_H
#define COLLECTOR_H
#include<utility/errorclass.h>
#include"gdat.h"
#include"dmap.h"
#include<RACE_LogfileBase.h>
#include<RACE_Logger.h>
#include"LoggerSnapshotConnection.h"
#include"KVSnapshotDumpable.h"
// for the Container
#include<ace/RB_Tree.h>
#include<Gdatcmp.h>
#include<ace/Atomic_Op.h>
#include<RACE_SnapshotTrigger.h>
#include<RACE_Depot.h>

class Collector;
class KVSSnapshotEntry : public RACE_SnapshotQueue::Entry
{
public:
  KVSSnapshotEntry(time_t timestamp,int pending,
                  Collector *pcd,
                  KVSnapshotDumpable *snapshot,
                  RACE_Logger* logger)
    : RACE_SnapshotQueue::Entry(timestamp,pending),pcd_(pcd),
      snapshot_(snapshot),logger_(logger)
    {}
  virtual void Log();
private: 
  Collector *pcd_;
  KVSnapshotDumpable *snapshot_;
  RACE_Logger* logger_;
};

//##Documentation
//## collects key/Value Pairs - Snapshots , generates Dumpables
//## Note: for multiple Types we would separate the Collector in
//## a Collector and a Container for each types.
class Collector: public Dmap,
	         public LoggerSnapshotConnection,
	         public RACE_SnapshotCreator,
	         public CorrespondingCollector
{
public:
  Collector(const char *rumpfpfad,int genlog=0);
  //-- Interface Dmap
  virtual int open() throw(Databaseerror);
  virtual int close();
  virtual int fetch(Gdat &key,Gdat &val) throw(Databaseerror);
  virtual int store(Gdat &key,Gdat &value,Dmapflag flag) throw(Databaseerror);

  int store(Gdat &key,Timestamp_t timestamp,Gdat &value,Dmapflag flag) throw(Databaseerror);
  virtual void asciidump(std::ostream& str);
  virtual void Checkpoint();

  void Checkpoint(time_t timestamp);
  //-- Interface LoggerSnapshotConnection
  virtual void SetOffset(RACE_LogfileOffset offset,time_t timestamp);
  virtual void InformDiscard(time_t timestamp);
  virtual void KeyValueInsertCallback(const Gdat& key,RACE_LogfileOffset off);
  virtual void KeyValueDiscardCallback(const Gdat& key);

  //-- Interface 
  //##Documentation RACE_SnapshotCreator
  //## From RACE_SnapshotCreator
  virtual int GetPendingTotal();
  //##Documentation
  //## From RACE_SnapshotCreator
  virtual RACE_SnapshotQueue::Entry *GenNewEntry(time_t timestamp,int pending);

  int insert(KVDumpable *kvd);

  //##Documentation
  //## from Interface CorrspondingCollector
  void RewriteAllAndClear();

  //-- Referencecounting - may be factored out in a parentclass
  void IncrPending();
  void DecrPending();
  void IncrSnapshotPending();
  void DecrSnapshotPending();
  virtual void release();
  //##Documentation
  //## inhibit deletion to early during flush
  void IncrWorking();
  //## Documentation
  //## Delete if not longer needed
  void DecrWorking();
  //##Documentation
  //## Check if we are during flush
  int GetWorking() { return working_.value(); }


protected:
  virtual ~Collector();
  virtual bool CheckRightToExist();
  //##Documentation
  //## stor internaly in the tree
  int store_i(Gdat &key,Timestamp_t timestamp,Gdat &value,Dmapflag flag) throw(Databaseerror);
  char rumpfpfad_[128];
  int genlog_;
  RACE_LogfileBase *logfile_;
  RACE_Logger *logger_;

private:
  struct TreeEntry_t
  {
    Gdat key;
    Timestamp_t timestamp;
    //##Documentation
    //## The value - for optimization kept in Memory here
    Gdat value;
    //##Documentation
    //## The entry in the KVSnapshotDumpable
    Logfile::KeyRecord *entry;

    TreeEntry_t(const Gdat &pkey,Timestamp_t t,const Gdat &pval):key(pkey),timestamp(t),value(pval),entry(0) {}
  };
  typedef ACE_RB_Tree<Gdat,TreeEntry_t,GdatCmp,ACE_Null_Mutex> KVTreeImpl;
  KVTreeImpl kvtree_;
  KVSnapshotDumpable *snapshotdumpable_;
  ::RACE_SnapshotTrigger snapshottrigger_;
  //##Documentation
  //## We may called from 2 threads:
  //## - The Thread calls Store
  //## - The Active Loggerthread calls SetOffset
  ACE_Thread_Mutex mutex_;
  //##Documentation 
  //## if dirty=false, no changes are made
  bool dirty_; 
  //##Documentation
  //## in case of a Depot, the depot wants to be informed when an snapshot
  //## is written. 
  //## in the Logfile case, it is null
  RACE_Depot *informsnapshot_;
  //##Documentation
  //## Number of pending CloneDumpables
  ACE_Atomic_Op<ACE_Thread_Mutex, long> pending_;
  //##Documentation
  //## Number of pending CloneSnapshotDumpables
  ACE_Atomic_Op<ACE_Thread_Mutex, long> pendingsnapshots_;
  //##Documentation
  //## Marker which prohibits deleting during flushing an Accumulator
  ACE_Atomic_Op<ACE_Thread_Mutex, long> working_;
  ACE_Atomic_Op<ACE_Thread_Mutex, long> refcount_;
};

#endif

