#include"KVDumpable.h"

KVDumpable::KVDumpable(Gdat &key,Timestamp_t timestamp,Gdat &value)
  :RACE_Dumpable(sizeof(Logfile::KVEntry)+key.get_len()+value.get_len()),
   lsc_(NULL)
{
  Logfile::KVEntry *content = (Logfile::KVEntry *)base();
  content->h.len=size();
  content->h.elementtype=Logfile::KV_Type;
  content->timestamp=timestamp;
  content->keylen = key.get_len();
  content->vallen = value.get_len();
  wr_ptr(sizeof(Logfile::KVEntry));
  copy((char *)(key.get_data()),key.get_len());
  copy((char *)(value.get_data()),value.get_len());
}

KVDumpable::KVDumpable(RACE::DataBlock *db,bool frag,RACE_LogfileOffset lo)
  :RACE_Dumpable(db,frag,lo)
{
  RACE_COND_FAIL(frag,"%s:%d: Handling of fragmented Dumpables not yet implemented",__FILE__,__LINE__);
}

void KVDumpable::GetKey(Gdat &key)
{
  Logfile::KVEntry *entry=GetContent();
  key.set(entry->key,entry->keylen);
}
 
void KVDumpable::GetValue(Gdat &value)
{
  Logfile::KVEntry *entry=GetContent();
  value.set(entry->key+entry->keylen,entry->vallen);
}

Timestamp_t KVDumpable::GetTimestamp()
{
  return GetContent()->timestamp;
}

void KVDumpable::InformOffset(RACE_LogfileOffset offset)
{
  if(lsc_)
  {
    Gdat mykey;
    GetKey(mykey);
    lsc_->KeyValueInsertCallback(mykey,offset);
  }
}

void KVDumpable::Discard()
{
  if(lsc_)
  {
    Gdat mykey;
    GetKey(mykey);
    lsc_->KeyValueDiscardCallback(mykey);
  }
}

void KVDumpable::CallVisitor(RACE::DumpableVisitor *visitor)
{
  KVSemantik *sem=dynamic_cast<KVSemantik *>(visitor);
  if(sem)
  {
    sem->KeyValue(this);
  }
}

RACE_Message_Block *KVDumpable::CloneManagementData() const
{
  return new KVDumpable(0);
}

