/**************************************************************************
 * Definition....: gdat.h                                                 *
 * Zweck.........: Klasse Gdat generalisiert die Wertbehlter fr die     *
 *                 leichtgewichtige Datenbanken                           *
 *                 ohne viel C++ schnickschnack                           *
 * Copyright 1998 Informatik-und Netzwerkverein Ravensburg e.V            *
 *                nach GNU Copyleft                                       *
 * von:           Rudolf Weber                                            *
 * Ergaenzungen 2004:                                                      *
 *  * Anpassung an neue C++-Version                                       *
 *  * Default-Constructor,Copy-Construktor und Zuweisungsoperator         *
 *  * get_const_data() dem const-Qualifizierer zu liebe
 **************************************************************************/
 
#ifndef GDAT_H
#define GDAT_H
#include<utility/errorclass.h>
#ifdef WITH_DB_CXX
#include<db_cxx.h>
#endif
#ifdef WITH_DB
#include<db.h>
#endif
#ifndef _GDBM_H_
// Identisch aus ndbm.h und gdbm.h , sonst Konflikt
typedef struct {
	char *dptr;
	size_t dsize;
      } datum;
#endif

enum PSpeichertyp { PST_STATIC=0,PST_DYNAMIC=1 };

class Gdat
{
  char *umem;      // Speicher vom aufrufer
  size_t  ulen;       // Lnge desselben
  char *allocmem;  // mit malloc allocierter Speicher 
  size_t alen;        // Lnge desselben
  size_t len;         // Lnge der Daten
  
public:
  Gdat();
  Gdat(char *usermem,size_t usermemlen);
  Gdat(const Gdat &other) throw(Error);
  Gdat& Gdat::operator=(const Gdat &other) throw(Error);
  ~Gdat() { frei(); }
  void frei();
  void set(char *usermem,size_t lang);
  void set(datum &d,PSpeichertyp st) throw(Error); // Fr GDBM, und ndbm
#ifdef WITH_DB_CXX
  void set(Dbt &d);          // fr berkleydb
#endif
#ifdef WITH_DB
  void set(DBT &d);          // holt wert
  void exportmem(DBT &erg);  // exportiert freien Speicher
  void export_val(DBT &erg); // exportiert Wert
#endif
  const void * get_const_data() const
    {
      return (const void *)(allocmem ? allocmem : umem);
    }
  void *get_data() { return (allocmem ? allocmem : umem); }
  size_t  get_len() const { return len; }
  size_t  get_ulen() const { return (allocmem ? alen : ulen); }
  void set_len(size_t lang) throw (Error); // zum krzen der Lnge

  void exchangetake(Gdat& v); // bernimmt den speicher oder tauscht usermem
  char *take()       // Nutzer verspricht allocmen freizugeben
    {
      char *ret=allocmem;
      allocmem=NULL; alen=0;
      return ret;   // falls NULL, war es usermem
    }
  // Garantiert das Speicher lnge l vorhanden ist
  void *garantie(size_t l) throw(Error);
};

#include<iostream>
std::ostream& operator<<(std::ostream& str,Gdat& d);
#endif


